#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers the Z-Y decomposition for an arbitrary one qubit gate.

See paper "Elementary gates for quantum computing" by Adriano Barenco et al.,
arXiv:quant-ph/9503016v1. (Note: They use different gate definitions!)
Or see theorem 4.1 in Nielsen and Chuang.

Decompose an arbitrary one qubit gate U into
U = e^(i alpha) Rz(beta) Ry(gamma) Rz(delta). If a gate V is element of SU(2),
i.e., determinant == 1, then
V = Rz(beta) Ry(gamma) Rz(delta)

"""

import cmath
import itertools
import math


import numpy

from projectq.cengines import DecompositionRule
from projectq.meta import Control, get_control_count
from projectq.ops import BasicGate, Ph, Ry, Rz


TOLERANCE = 1e-12


def _recognize_arb1qubit(cmd):
    """
    Recognize an arbitrary one qubit gate which has a matrix property.

    It does not allow gates which have control qubits as otherwise the
    AutoReplacer might go into an infinite loop. Use
    carb1qubit2cnotrzandry instead.
    """
    try:
        m = cmd.gate.matrix
        return (len(m) == 2 and get_control_count(cmd) == 0)
    except AttributeError:
        return False


def _test_parameters(matrix, a, b_half, c_half, d_half):
    """
    It builds matrix U with parameters (a, b/2, c/2, d/2) and compares against
    matrix.

    U = [[exp(j*(a-b/2-d/2))*cos(c/2), -exp(j*(a-b/2+d/2))*sin(c/2)],
         [exp(j*(a+b/2-d/2))*sin(c/2), exp(j*(a+b/2+d/2))*cos(c/2)]]

    Args:
        matrix(list): 2x2 matrix
        a: parameter of U
        b_half: b/2. parameter of U
        c_half: c/2. parameter of U
        d_half: d/2. parameter of U

    Returns:
        True if matrix elements of U and `matrix` are TOLERANCE close.
    """
    U = [[cmath.exp(1j*(a-b_half-d_half))*math.cos(c_half),
          -cmath.exp(1j*(a-b_half+d_half))*math.sin(c_half)],
         [cmath.exp(1j*(a+b_half-d_half))*math.sin(c_half),
          cmath.exp(1j*(a+b_half+d_half))*math.cos(c_half)]]
    return numpy.allclose(U, matrix, rtol=10*TOLERANCE, atol=TOLERANCE)


def _find_parameters(matrix):
    """
    Given a 2x2 unitary matrix, find the parameters
    a, b/2, c/2, and d/2 such that
    matrix == [[exp(j*(a-b/2-d/2))*cos(c/2), -exp(j*(a-b/2+d/2))*sin(c/2)],
               [exp(j*(a+b/2-d/2))*sin(c/2), exp(j*(a+b/2+d/2))*cos(c/2)]]

    Note:
    If the matrix is element of SU(2) (determinant == 1), then
    we can choose a = 0.

    Args:
        matrix(list): 2x2 unitary matrix

    Returns:
        parameters of the matrix: (a, b/2, c/2, d/2)
    """
    # Determine a, b/2, c/2 and d/2 (3 different cases).
    # Note: everything is modulo 2pi.
    # Case 1: sin(c/2) == 0:
    if abs(matrix[0][1]) < TOLERANCE:
        two_a = cmath.phase(matrix[0][0]*matrix[1][1]) % (2*math.pi)
        if abs(two_a) < TOLERANCE or abs(two_a) > 2*math.pi-TOLERANCE:
            # from 2a==0 (mod 2pi), it follows that a==0 or a==pi,
            # w.l.g. we can choose a==0 because (see U above)
            # c/2 -> c/2 + pi would have the same effect as as a==0 -> a==pi.
            a = 0
        else:
            a = two_a/2.
        d_half = 0  # w.l.g
        b = cmath.phase(matrix[1][1])-cmath.phase(matrix[0][0])
        possible_b_half = [(b/2.) % (2*math.pi), (b/2.+math.pi) % (2*math.pi)]
        # As we have fixed a, we need to find correct sign for cos(c/2)
        possible_c_half = [0.0, math.pi]
        found = False
        for b_half, c_half in itertools.product(possible_b_half,
                                                possible_c_half):
            if _test_parameters(matrix, a, b_half, c_half, d_half):
                found = True
                break
        if not found:
            raise Exception("Couldn't find parameters for matrix ", matrix,
                            "This shouldn't happen. Maybe the matrix is " +
                            "not unitary?")
    # Case 2: cos(c/2) == 0:
    elif abs(matrix[0][0]) < TOLERANCE:
        two_a = cmath.phase(-matrix[0][1]*matrix[1][0]) % (2*math.pi)
        if abs(two_a) < TOLERANCE or abs(two_a) > 2*math.pi-TOLERANCE:
            # from 2a==0 (mod 2pi), it follows that a==0 or a==pi,
            # w.l.g. we can choose a==0 because (see U above)
            # c/2 -> c/2 + pi would have the same effect as as a==0 -> a==pi.
            a = 0
        else:
            a = two_a/2.
        d_half = 0  # w.l.g
        b = cmath.phase(matrix[1][0])-cmath.phase(matrix[0][1]) + math.pi
        possible_b_half = [(b/2.) % (2*math.pi), (b/2.+math.pi) % (2*math.pi)]
        # As we have fixed a, we need to find correct sign for sin(c/2)
        possible_c_half = [math.pi/2., 3./2.*math.pi]
        found = False
        for b_half, c_half in itertools.product(possible_b_half,
                                                possible_c_half):
            if _test_parameters(matrix, a, b_half, c_half, d_half):
                found = True
                break
        if not found:
            raise Exception("Couldn't find parameters for matrix ", matrix,
                            "This shouldn't happen. Maybe the matrix is " +
                            "not unitary?")
    # Case 3: sin(c/2) != 0 and cos(c/2) !=0:
    else:
        two_a = cmath.phase(matrix[0][0]*matrix[1][1]) % (2*math.pi)
        if abs(two_a) < TOLERANCE or abs(two_a) > 2*math.pi-TOLERANCE:
            # from 2a==0 (mod 2pi), it follows that a==0 or a==pi,
            # w.l.g. we can choose a==0 because (see U above)
            # c/2 -> c/2 + pi would have the same effect as as a==0 -> a==pi.
            a = 0
        else:
            a = two_a/2.
        two_d = 2.*cmath.phase(matrix[0][1])-2.*cmath.phase(matrix[0][0])
        possible_d_half = [two_d/4. % (2*math.pi),
                           (two_d/4.+math.pi/2.) % (2*math.pi),
                           (two_d/4.+math.pi) % (2*math.pi),
                           (two_d/4.+3./2.*math.pi) % (2*math.pi)]
        two_b = 2.*cmath.phase(matrix[1][0])-2.*cmath.phase(matrix[0][0])
        possible_b_half = [two_b/4. % (2*math.pi),
                           (two_b/4.+math.pi/2.) % (2*math.pi),
                           (two_b/4.+math.pi) % (2*math.pi),
                           (two_b/4.+3./2.*math.pi) % (2*math.pi)]
        tmp = math.acos(abs(matrix[1][1]))
        possible_c_half = [tmp % (2*math.pi),
                           (tmp+math.pi) % (2*math.pi),
                           (-1.*tmp) % (2*math.pi),
                           (-1.*tmp+math.pi) % (2*math.pi)]
        found = False
        for b_half, c_half, d_half in itertools.product(possible_b_half,
                                                        possible_c_half,
                                                        possible_d_half):
            if _test_parameters(matrix, a, b_half, c_half, d_half):
                found = True
                break
        if not found:
            raise Exception("Couldn't find parameters for matrix ", matrix,
                            "This shouldn't happen. Maybe the matrix is " +
                            "not unitary?")
    return (a, b_half, c_half, d_half)


def _decompose_arb1qubit(cmd):
    """
    Use Z-Y decomposition of Nielsen and Chuang (Theorem 4.1).

    An arbitrary one qubit gate matrix can be writen as
    U = [[exp(j*(a-b/2-d/2))*cos(c/2), -exp(j*(a-b/2+d/2))*sin(c/2)],
         [exp(j*(a+b/2-d/2))*sin(c/2), exp(j*(a+b/2+d/2))*cos(c/2)]]
    where a,b,c,d are real numbers.
    Then U = exp(j*a) Rz(b) Ry(c) Rz(d).
    If the matrix is element of SU(2) (determinant == 1), then
    we can choose a = 0.
    """
    matrix = cmd.gate.matrix.tolist()
    a, b_half, c_half, d_half = _find_parameters(matrix)
    qb = cmd.qubits
    eng = cmd.engine
    with Control(eng, cmd.control_qubits):
        if Rz(2*d_half) != Rz(0):
            Rz(2*d_half) | qb
        if Ry(2*c_half) != Ry(0):
            Ry(2*c_half) | qb
        if Rz(2*b_half) != Rz(0):
            Rz(2*b_half) | qb
        if a != 0:
            Ph(a) | qb


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(BasicGate, _decompose_arb1qubit, _recognize_arb1qubit)
]



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Registers a decomposition for phase estimation.

(reference https://en.wikipedia.org/wiki/Quantum_phase_estimation_algorithm)

The Quantum Phase Estimation (QPE) executes the algorithm up to the inverse
QFT included. The following steps measuring the ancillas and computing the
phase should be executed outside of the QPE.

The decomposition uses as ancillas (qpe_ancillas) the first qubit/qureg in
the Command and as system qubits teh second qubit/qureg in the Command.

The unitary operator for which the phase estimation is estimated (unitary)
is the gate in Command

Example:
    .. code-block:: python

       # Example using a ProjectQ gate

       n_qpe_ancillas = 3
       qpe_ancillas = eng.allocate_qureg(n_qpe_ancillas)
       system_qubits = eng.allocate_qureg(1)
       angle = cmath.pi*2.*0.125
       U = Ph(angle) # unitary_specfic_to_the_problem()

       # Apply Quantum Phase Estimation
       QPE(U) | (qpe_ancillas, system_qubits)

       All(Measure) | qpe_ancillas
       # Compute the phase from the ancilla measurement
       #(https://en.wikipedia.org/wiki/Quantum_phase_estimation_algorithm)
       phasebinlist = [int(q) for q in qpe_ancillas]
       phase_in_bin = ''.join(str(j) for j in phasebinlist)
       phase_int = int(phase_in_bin,2)
       phase = phase_int / (2 ** n_qpe_ancillas)
       print (phase)

       # Example using a function (two_qubit_gate).
       # Instead of applying QPE on a gate U one could provide a function

       def two_qubit_gate(system_q, time):
           CNOT | (system_q[0], system_q[1])
           Ph(2.0*cmath.pi*(time * 0.125)) | system_q[1]
           CNOT | (system_q[0], system_q[1])

       n_qpe_ancillas = 3
       qpe_ancillas = eng.allocate_qureg(n_qpe_ancillas)
       system_qubits = eng.allocate_qureg(2)
       X | system_qubits[0]

       # Apply Quantum Phase Estimation
       QPE(two_qubit_gate) | (qpe_ancillas, system_qubits)

       All(Measure) | qpe_ancillas
       # Compute the phase from the ancilla measurement
       #(https://en.wikipedia.org/wiki/Quantum_phase_estimation_algorithm)
       phasebinlist = [int(q) for q in qpe_ancillas]
       phase_in_bin = ''.join(str(j) for j in phasebinlist)
       phase_int = int(phase_in_bin,2)
       phase = phase_int / (2 ** n_qpe_ancillas)
       print (phase)

Attributes:
    unitary (BasicGate): Unitary Operation either a ProjectQ gate or a
        function f.
        Calling the function with the parameters system_qubits(Qureg) and time
        (integer), i.e. f(system_qubits, time), applies to the system qubits a
        unitary defined in f with parameter time.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Control, Loop
from projectq.ops import H, Tensor, get_inverse, QFT

from projectq.ops import QPE


def _decompose_QPE(cmd):
    """ Decompose the Quantum Phase Estimation gate. """
    eng = cmd.engine

    # Ancillas is the first qubit/qureg. System-qubit is the second qubit/qureg
    qpe_ancillas = cmd.qubits[0]
    system_qubits = cmd.qubits[1]

    # Hadamard on the ancillas
    Tensor(H) | qpe_ancillas

    # The Unitary Operator
    U = cmd.gate.unitary

    # Control U on the system_qubits
    if (callable(U)):
        # If U is a function
        for i in range(len(qpe_ancillas)):
            with Control(eng, qpe_ancillas[i]):
                U(system_qubits, time=2**i)
    else:
        for i in range(len(qpe_ancillas)):
            ipower = int(2**i)
            with Loop(eng, ipower):
                with Control(eng, qpe_ancillas[i]):
                    U | system_qubits

    # Inverse QFT on the ancillas
    get_inverse(QFT) | qpe_ancillas


#: Decomposition rules
all_defined_decomposition_rules = [DecompositionRule(QPE, _decompose_QPE)]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from projectq.ops import BasicMathGate


class AddConstant(BasicMathGate):
    """
    Add a constant to a quantum number represented by a quantum register,
    stored from low- to high-bit.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[1] # qunum is now equal to 2
            AddConstant(3) | qunum # qunum is now equal to 5
    """
    def __init__(self, a):
        """
        Initializes the gate to the number to add.

        Args:
            a (int): Number to add to a quantum register.

        It also initializes its base class, BasicMathGate, with the
        corresponding function, so it can be emulated efficiently.
        """
        BasicMathGate.__init__(self, lambda x: ((x + a),))
        self.a = a

    def get_inverse(self):
        """
        Return the inverse gate (subtraction of the same constant).
        """
        return SubConstant(self.a)

    def __str__(self):
        return "AddConstant({})".format(self.a)

    def __eq__(self, other):
        return isinstance(other, AddConstant) and self.a == other.a

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)


def SubConstant(a):
    """
    Subtract a constant from a quantum number represented by a quantum
    register, stored from low- to high-bit.

    Args:
        a (int): Constant to subtract

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[2] # qunum is now equal to 4
            SubConstant(3) | qunum # qunum is now equal to 1
    """
    return AddConstant(-a)


class AddConstantModN(BasicMathGate):
    """
    Add a constant to a quantum number represented by a quantum register
    modulo N.

    The number is stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[1] # qunum is now equal to 2
            AddConstantModN(3, 4) | qunum # qunum is now equal to 1

    .. note::

      Pre-conditions:

      * c < N
      * c >= 0
      * The value stored in the quantum register must be lower than N
    """
    def __init__(self, a, N):
        """
        Initializes the gate to the number to add modulo N.

        Args:
            a (int): Number to add to a quantum register (0 <= a < N).
            N (int): Number modulo which the addition is carried out.

        It also initializes its base class, BasicMathGate, with the
        corresponding function, so it can be emulated efficiently.
        """
        BasicMathGate.__init__(self, lambda x: ((x + a) % N,))
        self.a = a
        self.N = N

    def __str__(self):
        return "AddConstantModN({}, {})".format(self.a, self.N)

    def get_inverse(self):
        """
        Return the inverse gate (subtraction of the same number a modulo the
        same number N).
        """
        return SubConstantModN(self.a, self.N)

    def __eq__(self, other):
        return (isinstance(other, AddConstantModN) and self.a == other.a and
                self.N == other.N)

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)


def SubConstantModN(a, N):
    """
    Subtract a constant from a quantum number represented by a quantum
    register modulo N.

    The number is stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Args:
        a (int): Constant to add
        N (int): Constant modulo which the addition of a should be carried
            out.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(3) # 3-qubit number
            X | qunum[1] # qunum is now equal to 2
            SubConstantModN(4,5) | qunum # qunum is now -2 = 6 = 1 (mod 5)

    .. note::

      Pre-conditions:

      * c < N
      * c >= 0
      * The value stored in the quantum register must be lower than N
    """
    return AddConstantModN(N - a, N)


class MultiplyByConstantModN(BasicMathGate):
    """
    Multiply a quantum number represented by a quantum register by a constant
    modulo N.

    The number is stored from low- to high-bit, i.e., qunum[0] is the LSB.

    Example:
        .. code-block:: python

            qunum = eng.allocate_qureg(5) # 5-qubit number
            X | qunum[2] # qunum is now equal to 4
            MultiplyByConstantModN(3,5) | qunum # qunum is now 2.

    .. note::

      Pre-conditions:

      * c < N
      * c >= 0
      * gcd(c, N) == 1
      * The value stored in the quantum register must be lower than N
    """
    def __init__(self, a, N):
        """
        Initializes the gate to the number to multiply with modulo N.

        Args:
            a (int): Number by which to multiply a quantum register
                (0 <= a < N).
            N (int): Number modulo which the multiplication is carried out.

        It also initializes its base class, BasicMathGate, with the
        corresponding function, so it can be emulated efficiently.
        """
        BasicMathGate.__init__(self, lambda x: ((a * x) % N,))
        self.a = a
        self.N = N

    def __str__(self):
        return "MultiplyByConstantModN({}, {})".format(self.a, self.N)

    def __eq__(self, other):
        return (isinstance(other, MultiplyByConstantModN) and
                self.a == other.a and self.N == other.N)

    def __hash__(self):
        return hash(str(self))

    def __ne__(self, other):
        return not self.__eq__(other)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains a compiler engine to map to the 5-qubit IBM chip
"""
import itertools

from projectq.cengines import BasicMapperEngine
from projectq.ops import FlushGate, NOT, Allocate
from projectq.meta import get_control_count
from projectq.backends import IBMBackend


class IBM5QubitMapper(BasicMapperEngine):
    """
    Mapper for the 5-qubit IBM backend.

    Maps a given circuit to the IBM Quantum Experience chip.

    Note:
        The mapper has to be run once on the entire circuit.

    Warning:
        If the provided circuit cannot be mapped to the hardware layout
        without performing Swaps, the mapping procedure
        **raises an Exception**.
    """
    def __init__(self, connections=None):
        """
        Initialize an IBM 5-qubit mapper compiler engine.

        Resets the mapping.
        """
        BasicMapperEngine.__init__(self)
        self.current_mapping = dict()
        self._reset()
        self._cmds = []
        self._interactions = dict()

        if connections is None:
            # General connectivity easier for testing functions
            self.connections = set([(0, 1), (1, 0), (1, 2), (1, 3), (1, 4),
                                    (2, 1), (2, 3), (2, 4), (3, 1), (3, 4),
                                    (4, 3)])
        else:
            self.connections = connections

    def is_available(self, cmd):
        """
        Check if the IBM backend can perform the Command cmd and return True
        if so.

        Args:
            cmd (Command): The command to check
        """
        return IBMBackend().is_available(cmd)

    def _reset(self):
        """
        Reset the mapping parameters so the next circuit can be mapped.
        """
        self._cmds = []
        self._interactions = dict()

    def _determine_cost(self, mapping):
        """
        Determines the cost of the circuit with the given mapping.

        Args:
            mapping (dict): Dictionary with key, value pairs where keys are
                logical qubit ids and the corresponding value is the physical
                location on the IBM Q chip.
        Returns:
            Cost measure taking into account CNOT directionality or None
            if the circuit cannot be executed given the mapping.
        """

        cost = 0
        for tpl in self._interactions:
            ctrl_id = tpl[0]
            target_id = tpl[1]
            ctrl_pos = mapping[ctrl_id]
            target_pos = mapping[target_id]
            if not (ctrl_pos, target_pos) in self.connections:
                if (target_pos, ctrl_pos) in self.connections:
                    cost += self._interactions[tpl]
                else:
                    return None
        return cost

    def _run(self):
        """
        Runs all stored gates.

        Raises:
            Exception:
                If the mapping to the IBM backend cannot be performed or if
                the mapping was already determined but more CNOTs get sent
                down the pipeline.
        """
        if (len(self.current_mapping) > 0
                and max(self.current_mapping.values()) > 4):
            raise RuntimeError("Too many qubits allocated. The IBM Q "
                               "device supports at most 5 qubits and no "
                               "intermediate measurements / "
                               "reallocations.")
        if len(self._interactions) > 0:
            logical_ids = list(self.current_mapping)
            best_mapping = self.current_mapping
            best_cost = None
            for physical_ids in itertools.permutations(list(range(5)),
                                                       len(logical_ids)):
                mapping = {
                    logical_ids[i]: physical_ids[i]
                    for i in range(len(logical_ids))
                }
                new_cost = self._determine_cost(mapping)
                if new_cost is not None:
                    if best_cost is None or new_cost < best_cost:
                        best_cost = new_cost
                        best_mapping = mapping
            if best_cost is None:
                raise RuntimeError("Circuit cannot be mapped without using "
                                   "Swaps. Mapping failed.")
            self._interactions = dict()
            self.current_mapping = best_mapping

        for cmd in self._cmds:
            self._send_cmd_with_mapped_ids(cmd)

        self._cmds = []

    def _store(self, cmd):
        """
        Store a command and handle CNOTs.

        Args:
            cmd (Command): A command to store
        """
        if not cmd.gate == FlushGate():
            target = cmd.qubits[0][0].id
        if _is_cnot(cmd):
            # CNOT encountered
            ctrl = cmd.control_qubits[0].id
            if not (ctrl, target) in self._interactions:
                self._interactions[(ctrl, target)] = 0
            self._interactions[(ctrl, target)] += 1
        elif cmd.gate == Allocate:
            if target not in self.current_mapping:
                new_max = 0
                if len(self.current_mapping) > 0:
                    new_max = max(self.current_mapping.values()) + 1
                self._current_mapping[target] = new_max
        self._cmds.append(cmd)

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it until
        completion.

        Args:
            command_list (list of Command objects): list of commands to
                receive.

        Raises:
            Exception: If mapping the CNOT gates to 1 qubit would require
                Swaps. The current version only supports remapping of CNOT
                gates without performing any Swaps due to the large costs
                associated with Swapping given the CNOT constraints.
        """
        for cmd in command_list:
            self._store(cmd)
            if isinstance(cmd.gate, FlushGate):
                self._run()
                self._reset()


def _is_cnot(cmd):
    """
    Check if the command corresponds to a CNOT (controlled NOT gate).

    Args:
        cmd (Command): Command to check whether it is a controlled NOT
            gate.
    """
    return (isinstance(cmd.gate, NOT.__class__)
            and get_control_count(cmd) == 1)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers the decomposition of an controlled arbitary single qubit gate.

See paper "Elementary gates for quantum computing" by Adriano Barenco et al.,
arXiv:quant-ph/9503016v1. (Note: They use different gate definitions!) or
Nielsen and Chuang chapter 4.3.
"""

import cmath
import itertools
import math

import numpy

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count, Control
from projectq.ops import BasicGate, Ph, Ry, Rz, X
from projectq.setups.decompositions import arb1qubit2rzandry as arb1q


TOLERANCE = 1e-12


def _recognize_carb1qubit(cmd):
    """ Recognize single controlled one qubit gates with a matrix."""
    if get_control_count(cmd) == 1:
        try:
            m = cmd.gate.matrix
            if len(m) == 2:
                return True
        except AttributeError:
            return False
    return False


def _test_parameters(matrix, a, b, c_half):
    """
    It builds matrix V with parameters (a, b, c/2) and compares against
    matrix.

    V = [[-sin(c/2) * exp(j*a), exp(j*(a-b)) * cos(c/2)],
         [exp(j*(a+b)) * cos(c/2), exp(j*a) * sin(c/2)]]

    Args:
        matrix(list): 2x2 matrix
        a: Parameter of V
        b: Parameter of V
        c_half: c/2. Parameter of V

    Returns:
        True if matrix elements of V and `matrix` are TOLERANCE close.
    """
    V = [[-math.sin(c_half)*cmath.exp(1j*a),
          cmath.exp(1j*(a-b))*math.cos(c_half)],
         [cmath.exp(1j*(a+b))*math.cos(c_half),
          cmath.exp(1j*a) * math.sin(c_half)]]
    return numpy.allclose(V, matrix, rtol=10*TOLERANCE, atol=TOLERANCE)


def _recognize_v(matrix):
    """
    Recognizes a matrix which can be written in the following form:

    V = [[-sin(c/2) * exp(j*a), exp(j*(a-b)) * cos(c/2)],
         [exp(j*(a+b)) * cos(c/2), exp(j*a) * sin(c/2)]]

    Args:
        matrix(list): 2x2 matrix
    Returns:
        False if it is not possible otherwise (a, b, c/2)
    """
    if abs(matrix[0][0]) < TOLERANCE:
        two_a = cmath.phase(matrix[0][1]*matrix[1][0]) % (2*math.pi)
        if abs(two_a) < TOLERANCE or abs(two_a) > 2*math.pi-TOLERANCE:
            # from 2a==0 (mod 2pi), it follows that a==0 or a==pi,
            # w.l.g. we can choose a==0 because (see U above)
            # c/2 -> c/2 + pi would have the same effect as as a==0 -> a==pi.
            a = 0
        else:
            a = two_a/2.
        two_b = cmath.phase(matrix[1][0])-cmath.phase(matrix[0][1])
        possible_b = [(two_b/2.) % (2*math.pi),
                      (two_b/2.+math.pi) % (2*math.pi)]
        possible_c_half = [0, math.pi]
        found = False
        for b, c_half in itertools.product(possible_b, possible_c_half):
            if _test_parameters(matrix, a, b, c_half):
                found = True
                break
        assert found  # It should work for all matrices with matrix[0][0]==0.
        return (a, b, c_half)

    if abs(matrix[0][1]) < TOLERANCE:
        two_a = cmath.phase(-matrix[0][0] * matrix[1][1]) % (2*math.pi)
        if abs(two_a) < TOLERANCE or abs(two_a) > 2*math.pi-TOLERANCE:
            # from 2a==0 (mod 2pi), it follows that a==0 or a==pi,
            # w.l.g. we can choose a==0 because (see U above)
            # c/2 -> c/2 + pi would have the same effect as as a==0 -> a==pi.
            a = 0
        else:
            a = two_a/2.
        b = 0
        possible_c_half = [math.pi/2., 3./2.*math.pi]
        found = False
        for c_half in possible_c_half:
            if _test_parameters(matrix, a, b, c_half):
                found = True
                return (a, b, c_half)
        return False

    two_a = cmath.phase(-1.*matrix[0][0]*matrix[1][1]) % (2*math.pi)
    if abs(two_a) < TOLERANCE or abs(two_a) > 2*math.pi-TOLERANCE:
        # from 2a==0 (mod 2pi), it follows that a==0 or a==pi,
        # w.l.g. we can choose a==0 because (see U above)
        # c/2 -> c/2 + pi would have the same effect as as a==0 -> a==pi.
        a = 0
    else:
        a = two_a/2.
    two_b = cmath.phase(matrix[1][0])-cmath.phase(matrix[0][1])
    possible_b = [(two_b/2.) % (2*math.pi),
                  (two_b/2.+math.pi) % (2*math.pi)]
    tmp = math.acos(abs(matrix[1][0]))
    possible_c_half = [tmp % (2*math.pi),
                       (tmp+math.pi) % (2*math.pi),
                       (-1.*tmp) % (2*math.pi),
                       (-1.*tmp+math.pi) % (2*math.pi)]
    found = False
    for b, c_half in itertools.product(possible_b, possible_c_half):
        if _test_parameters(matrix, a, b, c_half):
            found = True
            return (a, b, c_half)
    return False


def _decompose_carb1qubit(cmd):
    """
    Decompose the single controlled 1 qubit gate into CNOT, Rz, Ry, C(Ph).

    See Nielsen and Chuang chapter 4.3.

    An arbitrary one qubit gate matrix can be writen as
    U = [[exp(j*(a-b/2-d/2))*cos(c/2), -exp(j*(a-b/2+d/2))*sin(c/2)],
         [exp(j*(a+b/2-d/2))*sin(c/2), exp(j*(a+b/2+d/2))*cos(c/2)]]
    where a,b,c,d are real numbers.
    Then U = exp(j*a) Rz(b) Ry(c) Rz(d).

    Then C(U) = C(exp(ja)) * A * CNOT * B * CNOT * C with
    A = Rz(b) * Ry(c/2)
    B = Ry(-c/2) * Rz(-(d+b)/2)
    C = Rz((d-b)/2)
    Note that a == 0 if U is element of SU(2). Also note that
    the controlled phase C(exp(ia)) can be implemented with single
    qubit gates.

    If the one qubit gate matrix can be writen as
    V = [[-sin(c/2) * exp(j*a), exp(j*(a-b)) * cos(c/2)],
         [exp(j*(a+b)) * cos(c/2), exp(j*a) * sin(c/2)]]
    Then C(V) = C(exp(ja))* E * CNOT * D with
    E = Rz(b)Ry(c/2)
    D = Ry(-c/2)Rz(-b)
    This improvement is important for C(Y) or C(Z)

    For a proof follow Lemma 5.5 of Barenco et al.
    """
    matrix = cmd.gate.matrix.tolist()
    qb = cmd.qubits
    eng = cmd.engine

    # Case 1: Unitary matrix which can be written in the form of V:
    parameters_for_v = _recognize_v(matrix)
    if parameters_for_v:
        a, b, c_half = parameters_for_v
        if Rz(-b) != Rz(0):
            Rz(-b) | qb
        if Ry(-c_half) != Ry(0):
            Ry(-c_half) | qb
        with Control(eng, cmd.control_qubits):
            X | qb
        if Ry(c_half) != Ry(0):
            Ry(c_half) | qb
        if Rz(b) != Rz(0):
            Rz(b) | qb
        if a != 0:
            with Control(eng, cmd.control_qubits):
                Ph(a) | qb

    # Case 2: General matrix U:
    else:
        a, b_half, c_half, d_half = arb1q._find_parameters(matrix)
        d = 2*d_half
        b = 2*b_half
        if Rz((d-b)/2.) != Rz(0):
            Rz((d-b)/2.) | qb
        with Control(eng, cmd.control_qubits):
            X | qb
        if Rz(-(d+b)/2.) != Rz(0):
            Rz(-(d+b)/2.) | qb
        if Ry(-c_half) != Ry(0):
            Ry(-c_half) | qb
        with Control(eng, cmd.control_qubits):
            X | qb
        if Ry(c_half) != Ry(0):
            Ry(c_half) | qb
        if Rz(b) != Rz(0):
            Rz(b) | qb
        if a != 0:
            with Control(eng, cmd.control_qubits):
                Ph(a) | qb


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(BasicGate, _decompose_carb1qubit, _recognize_carb1qubit)
]



#   Copyright 2019 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Registers a decomposition for quantum amplitude amplification.

(Quick reference https://en.wikipedia.org/wiki/Amplitude_amplification.
Complete reference G. Brassard, P. Hoyer, M. Mosca, A. Tapp (2000)
Quantum Amplitude Amplification and Estimation
https://arxiv.org/abs/quant-ph/0005055)

Quantum Amplitude Amplification (QAA) executes the algorithm, but not
the final measurement required to obtain the marked state(s) with high
probability. The starting state on wich the QAA algorithm is executed
is the one resulting of aplying the Algorithm on the |0> state.

Example:
    .. code-block:: python

       def func_algorithm(eng,system_qubits):
           All(H) | system_qubits

       def func_oracle(eng,system_qubits,qaa_ancilla):
           # This oracle selects the state |010> as the one marked
           with Compute(eng):
              All(X) | system_qubits[0::2]
           with Control(eng, system_qubits):
              X | qaa_ancilla
           Uncompute(eng)

       system_qubits = eng.allocate_qureg(3)
       # Prepare the qaa_ancilla qubit in the |-> state
       qaa_ancilla = eng.allocate_qubit()
       X | qaa_ancilla
       H | qaa_ancilla

       # Creates the initial state form the Algorithm
       func_algorithm(eng, system_qubits)
       # Apply Quantum Amplitude Amplification the correct number of times
       num_it = int(math.pi/4.*math.sqrt(1 << 3))
       with Loop(eng, num_it):
         QAA(func_algorithm, func_oracle) | (system_qubits, qaa_ancilla)

       All(Measure) | system_qubits

Warning:
    No qubit allocation/deallocation may take place during the call
    to the defined Algorithm :code:`func_algorithm`

Attributes:
    func_algorithm: Algorithm that initialite the state and to be used
                    in the QAA algorithm
    func_oracle: The Oracle that marks the state(s) as "good"
    system_qubits: the system we are interested on
    qaa_ancilla: auxiliary qubit that helps to invert the amplitude of the
                 "good" states

"""

import math

from projectq.cengines import DecompositionRule
from projectq.meta import Control, Compute, CustomUncompute, Dagger
from projectq.ops import X, Z, Ph, All

from projectq.ops import QAA


def _decompose_QAA(cmd):
    """ Decompose the Quantum Amplitude Apmplification algorithm as a gate. """
    eng = cmd.engine

    # System-qubit is the first qubit/qureg. Ancilla qubit is the second qubit
    system_qubits = cmd.qubits[0]
    qaa_ancilla = cmd.qubits[1]

    # The Oracle and the Algorithm
    Oracle = cmd.gate.oracle
    A = cmd.gate.algorithm

    # Apply the oracle to invert the amplitude of the good states, S_Chi
    Oracle(eng, system_qubits, qaa_ancilla)

    # Apply the inversion of the Algorithm,
    # the inversion of the aplitude of |0> and the Algorithm

    with Compute(eng):
        with Dagger(eng):
            A(eng, system_qubits)
        All(X) | system_qubits
    with Control(eng, system_qubits[0:-1]):
        Z | system_qubits[-1]
    with CustomUncompute(eng):
        All(X) | system_qubits
        A(eng, system_qubits)
    Ph(math.pi) | system_qubits[0]


#: Decomposition rules
all_defined_decomposition_rules = [DecompositionRule(QAA, _decompose_QAA)]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a compiler engine which flips the directionality of CNOTs according
to the given connectivity graph. It also translates Swap gates to CNOTs if
necessary.
"""
from copy import deepcopy

from projectq.cengines import (BasicEngine,
                               ForwarderEngine,
                               CommandModifier)
from projectq.meta import get_control_count
from projectq.ops import (All,
                          NOT,
                          CNOT,
                          H,
                          Swap)


class SwapAndCNOTFlipper(BasicEngine):
    """
    Flips CNOTs and translates Swaps to CNOTs where necessary.

    Warning:
        This engine assumes that CNOT and Hadamard gates are supported by
        the following engines.

    Warning:
        This engine cannot be used as a backend.
    """

    def __init__(self, connectivity):
        """
        Initialize the engine.

        Args:
            connectivity (set): Set of tuples (c, t) where if (c, t) is an
                element of the set means that a CNOT can be performed between
                the physical ids (c, t) with c being the control and t being
                the target qubit.
        """
        BasicEngine.__init__(self)
        self.connectivity = connectivity

    def is_available(self, cmd):
        """
        Check if the IBM backend can perform the Command cmd and return True
        if so.

        Args:
            cmd (Command): The command to check
        """
        return self._is_swap(cmd) or self.next_engine.is_available(cmd)

    def _is_cnot(self, cmd):
        """
        Check if the command corresponds to a CNOT (controlled NOT gate).

        Args:
            cmd (Command): Command to check
        """
        return (isinstance(cmd.gate, NOT.__class__) and
                get_control_count(cmd) == 1)

    def _is_swap(self, cmd):
        """
        Check if the command corresponds to a Swap gate.

        Args:
            cmd (Command): Command to check
        """
        return (get_control_count(cmd) == 0 and cmd.gate == Swap)

    def _needs_flipping(self, cmd):
        """
        Return True if cmd is a CNOT which needs to be flipped around.

        Args:
            cmd (Command): Command to check
        """
        if not self._is_cnot(cmd):
            return False

        target = cmd.qubits[0][0].id
        control = cmd.control_qubits[0].id
        is_possible = (control, target) in self.connectivity
        if not is_possible and (target, control) not in self.connectivity:
            raise RuntimeError("The provided connectivity does not "
                               "allow to execute the CNOT gate {}."
                               .format(str(cmd)))
        return not is_possible

    def _send_cnot(self, cmd, control, target, flip=False):
        def cmd_mod(command):
            command.tags = deepcopy(cmd.tags) + command.tags
            command.engine = self.main_engine
            return command
        # We'll have to add all meta tags before sending on
        cmd_mod_eng = CommandModifier(cmd_mod)
        cmd_mod_eng.next_engine = self.next_engine
        cmd_mod_eng.main_engine = self.main_engine
        # forward everything to the command modifier
        forwarder_eng = ForwarderEngine(cmd_mod_eng)
        target[0].engine = forwarder_eng
        control[0].engine = forwarder_eng
        if flip:
            # flip the CNOT using Hadamard gates:
            All(H) | (control + target)
            CNOT | (target, control)
            All(H) | (control + target)
        else:
            CNOT | (control, target)

    def receive(self, command_list):
        """
        Receives a command list and if the command is a CNOT gate, it flips
        it using Hadamard gates if necessary; if it is a Swap gate, it
        decomposes it using 3 CNOTs. All other gates are simply sent to the
        next engine.

        Args:
            command_list (list of Command objects): list of commands to
                receive.
        """
        for cmd in command_list:
            if self._needs_flipping(cmd):
                self._send_cnot(cmd, cmd.control_qubits, cmd.qubits[0], True)
            elif self._is_swap(cmd):
                qubits = [qb for qr in cmd.qubits for qb in qr]
                ids = [qb.id for qb in qubits]
                assert len(ids) == 2
                if tuple(ids) in self.connectivity:
                    control = [qubits[0]]
                    target = [qubits[1]]
                elif tuple(reversed(ids)) in self.connectivity:
                    control = [qubits[1]]
                    target = [qubits[0]]
                else:
                    raise RuntimeError("The provided connectivity does not "
                                       "allow to execute the Swap gate {}."
                                       .format(str(cmd)))
                self._send_cnot(cmd, control, target)
                self._send_cnot(cmd, target, control, True)
                self._send_cnot(cmd, control, target)
            else:
                self.next_engine.receive([cmd])



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Defines LogicalQubitIDTag to annotate a MeasureGate for mapped qubits.
"""


class LogicalQubitIDTag(object):
    """
    LogicalQubitIDTag for a mapped qubit to annotate a MeasureGate.

    Attributes:
        logical_qubit_id (int): Logical qubit id
    """
    def __init__(self, logical_qubit_id):
        self.logical_qubit_id = logical_qubit_id

    def __eq__(self, other):
        return (isinstance(other, LogicalQubitIDTag) and
                self.logical_qubit_id == other.logical_qubit_id)

    def __ne__(self, other):
        return not self.__eq__(other)



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Module uses ideas from "Basic circuit compilation techniques
#   for an ion-trap quantum machine" by Dmitri Maslov (2017) at
#   https://iopscience.iop.org/article/10.1088/1367-2630/aa5e47
"""
Registers a decomposition to for a CNOT gate in terms of Rxx, Rx and Ry gates.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count
from projectq.ops import Ph, Rxx, Ry, Rx, X
import math


def _decompose_cnot2rxx_M(cmd):
    """ Decompose CNOT gate into Rxx gate. """
    # Labelled 'M' for 'minus' because decomposition ends with a Ry(-pi/2)
    ctrl = cmd.control_qubits
    Ry(math.pi / 2) | ctrl[0]
    Ph(7 * math.pi / 4) | ctrl[0]
    Rx(-math.pi / 2) | ctrl[0]
    Rx(-math.pi / 2) | cmd.qubits[0][0]
    Rxx(math.pi / 2) | (ctrl[0], cmd.qubits[0][0])
    Ry(-1 * math.pi / 2) | ctrl[0]


def _decompose_cnot2rxx_P(cmd):
    """ Decompose CNOT gate into Rxx gate. """
    # Labelled 'P' for 'plus' because decomposition ends with a Ry(+pi/2)
    ctrl = cmd.control_qubits
    Ry(-math.pi / 2) | ctrl[0]
    Ph(math.pi / 4) | ctrl[0]
    Rx(-math.pi / 2) | ctrl[0]
    Rx(math.pi / 2) | cmd.qubits[0][0]
    Rxx(math.pi / 2) | (ctrl[0], cmd.qubits[0][0])
    Ry(math.pi / 2) | ctrl[0]


def _recognize_cnot2(cmd):
    """ Identify that the command is a CNOT gate (control - X gate)"""
    return get_control_count(cmd) == 1


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(X.__class__, _decompose_cnot2rxx_M, _recognize_cnot2,
                      10),
    DecompositionRule(X.__class__, _decompose_cnot2rxx_P, _recognize_cnot2,
                      10)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition rule for multi-controlled gates.

Implements the decomposition of Nielsen and Chuang (Fig. 4.10) which
decomposes a C^n(U) gate into a sequence of 2 * (n-1) Toffoli gates and one
C(U) gate by using (n-1) ancilla qubits and circuit depth of 2n-1.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count, Compute, Control, Uncompute
from projectq.ops import BasicGate, Toffoli, XGate


def _recognize_CnU(cmd):
    """
    Recognize an arbitrary gate which has n>=2 control qubits, except a
    Toffoli gate.
    """
    if get_control_count(cmd) == 2:
        if not isinstance(cmd.gate, XGate):
            return True
    elif get_control_count(cmd) > 2:
        return True
    return False


def _decompose_CnU(cmd):
    """
    Decompose a multi-controlled gate U with n control qubits into a single-
    controlled U.

    It uses (n-1) work qubits and 2 * (n-1) Toffoli gates for general U
    and (n-2) work qubits and 2n - 3 Toffoli gates if U is an X-gate.
    """
    eng = cmd.engine
    qubits = cmd.qubits
    ctrl_qureg = cmd.control_qubits
    gate = cmd.gate
    n = get_control_count(cmd)

    # specialized for X-gate
    if gate == XGate() and n > 2:
        n -= 1
    ancilla_qureg = eng.allocate_qureg(n-1)

    with Compute(eng):
        Toffoli | (ctrl_qureg[0], ctrl_qureg[1], ancilla_qureg[0])
        for ctrl_index in range(2, n):
            Toffoli | (ctrl_qureg[ctrl_index], ancilla_qureg[ctrl_index-2],
                       ancilla_qureg[ctrl_index-1])
    ctrls = [ancilla_qureg[-1]]

    # specialized for X-gate
    if gate == XGate() and get_control_count(cmd) > 2:
        ctrls += [ctrl_qureg[-1]]
    with Control(eng, ctrls):
        gate | qubits

    Uncompute(eng)


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(BasicGate, _decompose_CnU, _recognize_CnU)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
This file defines BasicQubit, Qubit, WeakQubit and Qureg.

A Qureg represents a list of Qubit or WeakQubit objects.
Qubit represents a (logical-level) qubit with a unique index provided by the
MainEngine. Qubit objects are automatically deallocated if they go out of
scope and intented to be used within Qureg objects in user code.

Example:
    .. code-block:: python

        from projectq import MainEngine
        eng = MainEngine()
        qubit = eng.allocate_qubit()

qubit is a Qureg of size 1 with one Qubit object which is deallocated once
qubit goes out of scope.

WeakQubit are used inside the Command object and are not automatically
deallocated.
"""


class BasicQubit(object):
    """
    BasicQubit objects represent qubits.

    They have an id and a reference to the owning engine.
    """
    def __init__(self, engine, idx):
        """
        Initialize a BasicQubit object.

        Args:
            engine: Owning engine / engine that created the qubit
            idx: Unique index of the qubit referenced by this qubit
        """
        self.id = idx
        self.engine = engine

    def __str__(self):
        """
        Return string representation of this qubit.
        """
        return str(self.id)

    def __bool__(self):
        """
        Access the result of a previous measurement and return False / True
        (0 / 1)
        """
        return self.engine.main_engine.get_measurement_result(self)

    def __nonzero__(self):
        """
        Access the result of a previous measurement for Python 2.7.
        """
        return self.__bool__()

    def __int__(self):
        """
        Access the result of a previous measurement and return as integer
        (0 / 1).
        """
        return int(bool(self))

    def __eq__(self, other):
        """
        Compare with other qubit (Returns True if equal id and engine).

        Args:
            other (BasicQubit): BasicQubit to which to compare this one
        """
        if self.id == -1:
            return self is other
        return (isinstance(other, BasicQubit) and
                self.id == other.id and
                self.engine == other.engine)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        """
        Return the hash of this qubit.

        Hash definition because of custom __eq__.
        Enables storing a qubit in, e.g., a set.
        """
        if self.id == -1:
            return object.__hash__(self)
        return hash((self.engine, self.id))


class Qubit(BasicQubit):
    """
    Qubit class.

    Represents a (logical-level) qubit with a unique index provided by the
    MainEngine. Once the qubit goes out of scope (and is garbage-collected),
    it deallocates itself automatically, allowing automatic resource
    management.

    Thus the qubit is not copyable; only returns a reference to the same
    object.
    """
    def __del__(self):
        """
        Destroy the qubit and deallocate it (automatically).
        """
        if self.id == -1:
            return
        # If a user directly calls this function, then the qubit gets id == -1
        # but stays in active_qubits as it is not yet deleted, hence remove
        # it manually (if the garbage collector calls this function, then the
        # WeakRef in active qubits is already gone):
        if self in self.engine.main_engine.active_qubits:
            self.engine.main_engine.active_qubits.remove(self)
        weak_copy = WeakQubitRef(self.engine, self.id)
        self.id = -1
        self.engine.deallocate_qubit(weak_copy)

    def __copy__(self):
        """
        Non-copyable (returns reference to self).

        Note:
            To prevent problems with automatic deallocation, qubits are not
            copyable!
        """
        return self

    def __deepcopy__(self, memo):
        """
        Non-deepcopyable (returns reference to self).

        Note:
            To prevent problems with automatic deallocation, qubits are not
            deepcopyable!
        """
        return self


class WeakQubitRef(BasicQubit):
    """
    WeakQubitRef objects are used inside the Command object.

    Qubits feature automatic deallocation when destroyed. WeakQubitRefs, on
    the other hand, do not share this feature, allowing to copy them and pass
    them along the compiler pipeline, while the actual qubit objects may be
    garbage-collected (and, thus, cleaned up early). Otherwise there is no
    difference between a WeakQubitRef and a Qubit object.
    """
    pass


class Qureg(list):
    """
    Quantum register class.

    Simplifies accessing measured values for single-qubit registers (no []-
    access necessary) and enables pretty-printing of general quantum registers
    (call Qureg.__str__(qureg)).
    """
    def __bool__(self):
        """
        Return measured value if Qureg consists of 1 qubit only.

        Raises:
            Exception if more than 1 qubit resides in this register (then you
            need to specify which value to get using qureg[???])
        """
        if len(self) == 1:
            return bool(self[0])

        raise Exception("__bool__(qureg): Quantum register contains more "
                        "than 1 qubit. Use __bool__(qureg[idx]) instead.")

    def __int__(self):
        """
        Return measured value if Qureg consists of 1 qubit only.

        Raises:
            Exception if more than 1 qubit resides in this register (then you
            need to specify which value to get using qureg[???])
        """
        if len(self) == 1:
            return int(self[0])

        raise Exception("__int__(qureg): Quantum register contains more "
                        "than 1 qubit. Use __bool__(qureg[idx]) instead.")

    def __nonzero__(self):
        """
        Return measured value if Qureg consists of 1 qubit only for Python 2.7.

        Raises:
            Exception if more than 1 qubit resides in this register (then you
            need to specify which value to get using qureg[???])
        """
        return int(self) != 0

    def __str__(self):
        """
        Get string representation of a quantum register.
        """
        if len(self) == 0:
            return "Qureg[]"

        ids = [q.id for q in self[1:]]
        ids.append(None)  # Forces a flush on last loop iteration.

        out_list = []
        start_id = self[0].id
        count = 1
        for qubit_id in ids:
            if qubit_id == start_id + count:
                count += 1
                continue

            out_list.append('{}-{}'.format(start_id, start_id + count - 1)
                            if count > 1
                            else '{}'.format(start_id))
            start_id = qubit_id
            count = 1

        return "Qureg[{}]".format(', '.join(out_list))

    @property
    def engine(self):
        """
        Return owning engine.
        """
        return self[0].engine

    @engine.setter
    def engine(self, eng):
        """
        Set owning engine.
        """
        for qb in self:
            qb.engine = eng



#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Registers a decomposition for general unitary gates
"""

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count
from projectq.ops import Ph, Ry, Rz, U


def _decompose_U(cmd):
    """ Decompose a general unitary to Rz and Ry gates. """
    gate = cmd.gate
    qubit = cmd.qubits[0][0]
    Rz(gate.delta) | qubit
    Ry(gate.gamma) | qubit
    Rz(gate.beta) | qubit
    Ph(gate.alpha) | qubit


def _recognize_UNoCtrl(cmd):
    """
    Decompose the gate only if the command represents a single qubit gate
    (if it is not part of a control gate).
    """
    return get_control_count(cmd) == 0


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(U, _decompose_U, _recognize_UNoCtrl)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# helpers to run the jsonified gate sequence on ibm quantum experience server
# api documentation does not exist and has to be deduced from the qiskit code
# source at: https://github.com/Qiskit/qiskit-ibmq-provider

import getpass
import time
import signal
import requests
from requests.compat import urljoin
from requests import Session

_AUTH_API_URL = ('https://auth.quantum-computing.ibm.com/api/users/'
                 'loginWithToken')
_API_URL = 'https://api.quantum-computing.ibm.com/api/'

# TODO: call to get the API version automatically
CLIENT_APPLICATION = 'ibmqprovider/0.4.4'


class IBMQ(Session):
    """
    Manage a session between ProjectQ and the IBMQ web API.
    """

    def __init__(self, **kwargs):
        super(IBMQ, self).__init__(**kwargs)  # Python 2 compatibility
        self.backends = dict()
        self.timeout = 5.0

    def get_list_devices(self, verbose=False):
        """
        Get the list of available IBM backends with their properties

        Args:
            verbose (bool): print the returned dictionnary if True

        Returns:
            (dict) backends dictionary by name device, containing the qubit
                    size 'nq', the coupling map 'coupling_map' as well as the
                    device version 'version'
        """
        list_device_url = 'Network/ibm-q/Groups/open/Projects/main/devices/v/1'
        argument = {'allow_redirects': True, 'timeout': (self.timeout, None)}
        request = super(IBMQ, self).get(urljoin(_API_URL, list_device_url),
                                        **argument)
        request.raise_for_status()
        r_json = request.json()
        self.backends = dict()
        for el in r_json:
            self.backends[el['backend_name']] = {
                'nq': el['n_qubits'],
                'coupling_map': el['coupling_map'],
                'version': el['backend_version']
            }

        if verbose:
            print('- List of IBMQ devices available:')
            print(self.backends)
        return self.backends

    def is_online(self, device):
        """
        Check if the device is in the list of available IBM backends.

        Args:
            device (str): name of the device to check

        Returns:
            (bool) True if device is available, False otherwise
        """
        return device in self.backends

    def can_run_experiment(self, info, device):
        """
        Check if the device is big enough to run the code.

        Args:
            info (dict): dictionary sent by the backend containing the code to
                run
            device (str): name of the ibm device to use

        Returns:
            (tuple): (bool) True if device is big enough, False otherwise
                     (int) maximum number of qubit available on the device
                     (int) number of qubit needed for the circuit

        """
        nb_qubit_max = self.backends[device]['nq']
        nb_qubit_needed = info['nq']
        return nb_qubit_needed <= nb_qubit_max, nb_qubit_max, nb_qubit_needed

    def _authenticate(self, token=None):
        """
        Args:
            token (str): IBM quantum experience user API token.
        """
        if token is None:
            token = getpass.getpass(prompt="IBM QE token > ")
        if len(token) == 0:
            raise Exception('Error with the IBM QE token')
        self.headers.update({'X-Qx-Client-Application': CLIENT_APPLICATION})
        args = {
            'data': None,
            'json': {
                'apiToken': token
            },
            'timeout': (self.timeout, None)
        }
        request = super(IBMQ, self).post(_AUTH_API_URL, **args)
        request.raise_for_status()
        r_json = request.json()
        self.params.update({'access_token': r_json['id']})

    def _run(self, info, device):
        post_job_url = 'Network/ibm-q/Groups/open/Projects/main/Jobs'
        shots = info['shots']
        n_classical_reg = info['nq']
        n_qubits = self.backends[device]['nq']
        version = self.backends[device]['version']
        instructions = info['json']
        maxcredit = info['maxCredits']
        c_label = []
        q_label = []
        for i in range(n_classical_reg):
            c_label.append(['c', i])
        for i in range(n_qubits):
            q_label.append(['q', i])
        experiment = [{
            'header': {
                'qreg_sizes': [['q', n_qubits]],
                'n_qubits': n_qubits,
                'memory_slots': n_classical_reg,
                'creg_sizes': [['c', n_classical_reg]],
                'clbit_labels': c_label,
                'qubit_labels': q_label,
                'name': 'circuit0'
            },
            'config': {
                'n_qubits': n_qubits,
                'memory_slots': n_classical_reg
            },
            'instructions': instructions
        }]
        # Note: qobj_id is not necessary in projectQ, so fixed string for now
        argument = {
            'data': None,
            'json': {
                'qObject': {
                    'type': 'QASM',
                    'schema_version': '1.1.0',
                    'config': {
                        'shots': shots,
                        'max_credits': maxcredit,
                        'n_qubits': n_qubits,
                        'memory_slots': n_classical_reg,
                        'memory': False,
                        'parameter_binds': []
                    },
                    'experiments': experiment,
                    'header': {
                        'backend_version': version,
                        'backend_name': device
                    },
                    'qobj_id': 'e72443f5-7752-4e32-9ac8-156f1f3fee18'
                },
                'backend': {
                    'name': device
                },
                'shots': shots
            },
            'timeout': (self.timeout, None)
        }
        request = super(IBMQ, self).post(urljoin(_API_URL, post_job_url),
                                         **argument)
        request.raise_for_status()
        r_json = request.json()
        execution_id = r_json["id"]
        return execution_id

    def _get_result(self,
                    device,
                    execution_id,
                    num_retries=3000,
                    interval=1,
                    verbose=False):

        job_status_url = ('Network/ibm-q/Groups/open/Projects/main/Jobs/'
                          + execution_id)

        if verbose:
            print("Waiting for results. [Job ID: {}]".format(execution_id))

        original_sigint_handler = signal.getsignal(signal.SIGINT)

        def _handle_sigint_during_get_result(*_):
            raise Exception(
                "Interrupted. The ID of your submitted job is {}.".format(
                    execution_id))

        try:
            signal.signal(signal.SIGINT, _handle_sigint_during_get_result)
            for retries in range(num_retries):

                argument = {
                    'allow_redirects': True,
                    'timeout': (self.timeout, None)
                }
                request = super(IBMQ,
                                self).get(urljoin(_API_URL, job_status_url),
                                          **argument)
                request.raise_for_status()
                r_json = request.json()
                if r_json['status'] == 'COMPLETED':
                    return r_json['qObjectResult']['results'][0]
                if r_json['status'] != 'RUNNING':
                    raise Exception("Error while running the code: {}.".format(
                        r_json['status']))
                time.sleep(interval)
                if self.is_online(device) and retries % 60 == 0:
                    self.get_list_devices()
                    if not self.is_online(device):
                        raise DeviceOfflineError(
                            "Device went offline. The ID of "
                            "your submitted job is {}.".format(execution_id))

        finally:
            if original_sigint_handler is not None:
                signal.signal(signal.SIGINT, original_sigint_handler)

        raise Exception("Timeout. The ID of your submitted job is {}.".format(
            execution_id))


class DeviceTooSmall(Exception):
    pass


class DeviceOfflineError(Exception):
    pass


def show_devices(token=None, verbose=False):
    """
    Access the list of available devices and their properties (ex: for setup
    configuration)

    Args:
        token (str): IBM quantum experience user API token.
        verbose (bool): If True, additional information is printed

    Returns:
        (list) list of available devices and their properties
    """
    ibmq_session = IBMQ()
    ibmq_session._authenticate(token=token)
    return ibmq_session.get_list_devices(verbose=verbose)


def retrieve(device,
             token,
             jobid,
             num_retries=3000,
             interval=1,
             verbose=False):
    """
    Retrieves a previously run job by its ID.

    Args:
        device (str): Device on which the code was run / is running.
        token (str): IBM quantum experience user API token.
        jobid (str): Id of the job to retrieve

    Returns:
        (dict) result form the IBMQ server
    """
    ibmq_session = IBMQ()
    ibmq_session._authenticate(token)
    ibmq_session.get_list_devices(verbose)
    res = ibmq_session._get_result(device,
                                   jobid,
                                   num_retries=num_retries,
                                   interval=interval,
                                   verbose=verbose)
    return res


def send(info,
         device='ibmq_qasm_simulator',
         token=None,
         shots=None,
         num_retries=3000,
         interval=1,
         verbose=False):
    """
    Sends QASM through the IBM API and runs the quantum circuit.

    Args:
        info(dict): Contains representation of the circuit to run.
        device (str): name of the ibm device. Simulator chosen by default
        token (str): IBM quantum experience user API token.
        shots (int): Number of runs of the same circuit to collect statistics.
        verbose (bool): If True, additional information is printed, such as
            measurement statistics. Otherwise, the backend simply registers
            one measurement result (same behavior as the projectq Simulator).

    Returns:
        (dict) result form the IBMQ server

    """
    try:
        ibmq_session = IBMQ()
        # Shots argument deprecated, as already
        if shots is not None:
            info['shots'] = shots
        if verbose:
            print("- Authenticating...")
            if token is not None:
                print('user API token: ' + token)
        ibmq_session._authenticate(token)

        # check if the device is online
        ibmq_session.get_list_devices(verbose)
        online = ibmq_session.is_online(device)
        if not online:
            print("The device is offline (for maintenance?). Use the "
                  "simulator instead or try again later.")
            raise DeviceOfflineError("Device is offline.")

        # check if the device has enough qubit to run the code
        runnable, qmax, qneeded = ibmq_session.can_run_experiment(info, device)
        if not runnable:
            print(
                ("The device is too small ({} qubits available) for the code "
                 + "requested({} qubits needed) Try to look for another "
                 + "device with more qubits").format(qmax, qneeded))
            raise DeviceTooSmall("Device is too small.")
        if verbose:
            print("- Running code: {}".format(info))
        execution_id = ibmq_session._run(info, device)
        if verbose:
            print("- Waiting for results...")
        res = ibmq_session._get_result(device,
                                       execution_id,
                                       num_retries=num_retries,
                                       interval=interval,
                                       verbose=verbose)
        if verbose:
            print("- Done.")
        return res
    except requests.exceptions.HTTPError as err:
        print("- There was an error running your code:")
        print(err)
    except requests.exceptions.RequestException as err:
        print("- Looks like something is wrong with server:")
        print(err)
    except KeyError as err:
        print("- Failed to parse response:")
        print(err)



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition rule for a unitary QubitOperator to one qubit gates.
"""

import cmath

from projectq.cengines import DecompositionRule
from projectq.meta import Control, get_control_count
from projectq.ops import Ph, QubitOperator, X, Y, Z


def _recognize_qubitop(cmd):
    """ For efficiency only use this if at most 1 control qubit."""
    return get_control_count(cmd) <= 1


def _decompose_qubitop(cmd):
    assert len(cmd.qubits) == 1
    qureg = cmd.qubits[0]
    eng = cmd.engine
    qubit_op = cmd.gate
    with Control(eng, cmd.control_qubits):
        (term, coefficient), = qubit_op.terms.items()
        phase = cmath.phase(coefficient)
        Ph(phase) | qureg[0]
        for index, action in term:
            if action == "X":
                X | qureg[index]
            elif action == "Y":
                Y | qureg[index]
            elif action == "Z":
                Z | qureg[index]


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(QubitOperator, _decompose_qubitop, _recognize_qubitop)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a compiler engine which counts the number of calls for each type of
gate used in a circuit, in addition to the max. number of active qubits.
"""

from projectq.cengines import BasicEngine, LastEngineException
from projectq.meta import get_control_count, LogicalQubitIDTag
from projectq.ops import FlushGate, Deallocate, Allocate, Measure
from projectq.types import WeakQubitRef


class ResourceCounter(BasicEngine):
    """
    ResourceCounter is a compiler engine which counts the number of gates and
    max. number of active qubits.

    Attributes:
        gate_counts (dict): Dictionary of gate counts.
            The keys are tuples of the form (cmd.gate, ctrl_cnt), where
            ctrl_cnt is the number of control qubits.
        gate_class_counts (dict): Dictionary of gate class counts.
            The keys are tuples of the form (cmd.gate.__class__, ctrl_cnt),
            where ctrl_cnt is the number of control qubits.
        max_width (int): Maximal width (=max. number of active qubits at any
            given point).
    Properties:
        depth_of_dag (int): It is the longest path in the directed
                            acyclic graph (DAG) of the program.
    """
    def __init__(self):
        """
        Initialize a resource counter engine.

        Sets all statistics to zero.
        """
        BasicEngine.__init__(self)
        self.gate_counts = {}
        self.gate_class_counts = {}
        self._active_qubits = 0
        self.max_width = 0
        # key: qubit id, depth of this qubit
        self._depth_of_qubit = dict()
        self._previous_max_depth = 0

    def is_available(self, cmd):
        """
        Specialized implementation of is_available: Returns True if the
        ResourceCounter is the last engine (since it can count any command).

        Args:
            cmd (Command): Command for which to check availability (all
                Commands can be counted).

        Returns:
            availability (bool): True, unless the next engine cannot handle
                the Command (if there is a next engine).
        """
        try:
            return BasicEngine.is_available(self, cmd)
        except LastEngineException:
            return True

    @property
    def depth_of_dag(self):
        if self._depth_of_qubit:
            current_max = max(self._depth_of_qubit.values())
            return max(current_max, self._previous_max_depth)
        return self._previous_max_depth

    def _add_cmd(self, cmd):
        """
        Add a gate to the count.
        """
        if cmd.gate == Allocate:
            self._active_qubits += 1
            self._depth_of_qubit[cmd.qubits[0][0].id] = 0
        elif cmd.gate == Deallocate:
            self._active_qubits -= 1
            depth = self._depth_of_qubit[cmd.qubits[0][0].id]
            self._previous_max_depth = max(self._previous_max_depth, depth)
            self._depth_of_qubit.pop(cmd.qubits[0][0].id)
        elif self.is_last_engine and cmd.gate == Measure:
            for qureg in cmd.qubits:
                for qubit in qureg:
                    self._depth_of_qubit[qubit.id] += 1
                    # Check if a mapper assigned a different logical id
                    logical_id_tag = None
                    for tag in cmd.tags:
                        if isinstance(tag, LogicalQubitIDTag):
                            logical_id_tag = tag
                    if logical_id_tag is not None:
                        qubit = WeakQubitRef(qubit.engine,
                                             logical_id_tag.logical_qubit_id)
                    self.main_engine.set_measurement_result(qubit, 0)
        else:
            qubit_ids = set()
            for qureg in cmd.all_qubits:
                for qubit in qureg:
                    qubit_ids.add(qubit.id)
            if len(qubit_ids) == 1:
                self._depth_of_qubit[list(qubit_ids)[0]] += 1
            else:
                max_depth = 0
                for qubit_id in qubit_ids:
                    max_depth = max(max_depth, self._depth_of_qubit[qubit_id])
                for qubit_id in qubit_ids:
                    self._depth_of_qubit[qubit_id] = max_depth + 1

        self.max_width = max(self.max_width, self._active_qubits)

        ctrl_cnt = get_control_count(cmd)
        gate_description = (cmd.gate, ctrl_cnt)
        gate_class_description = (cmd.gate.__class__, ctrl_cnt)

        try:
            self.gate_counts[gate_description] += 1
        except KeyError:
            self.gate_counts[gate_description] = 1

        try:
            self.gate_class_counts[gate_class_description] += 1
        except KeyError:
            self.gate_class_counts[gate_class_description] = 1

    def __str__(self):
        """
        Return the string representation of this ResourceCounter.

        Returns:
            A summary (string) of resources used, including gates, number of
                calls, and max. number of qubits that were active at the same
                time.
        """
        if len(self.gate_counts) > 0:
            gate_class_list = []
            for gate_class_description, num in self.gate_class_counts.items():
                gate_class, ctrl_cnt = gate_class_description
                gate_class_name = ctrl_cnt * "C" + gate_class.__name__
                gate_class_list.append(gate_class_name + " : " + str(num))

            gate_list = []
            for gate_description, num in self.gate_counts.items():
                gate, ctrl_cnt = gate_description
                gate_name = ctrl_cnt * "C" + str(gate)
                gate_list.append(gate_name + " : " + str(num))

            return ("Gate class counts:\n    " +
                    "\n    ".join(list(sorted(gate_class_list))) +
                    "\n\nGate counts:\n    " +
                    "\n    ".join(list(sorted(gate_list))) +
                    "\n\nMax. width (number of qubits) : " +
                    str(self.max_width) + ".")
        return "(No quantum resources used)"

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine, increases the
        counters of the received commands, and then send them on to the next
        engine.

        Args:
            command_list (list<Command>): List of commands to receive (and
                count).
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self._add_cmd(cmd)

            # (try to) send on
            if not self.is_last_engine:
                self.send([cmd])



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""TestEngine and DummyEngine."""


from copy import deepcopy
from projectq.cengines import BasicEngine
from projectq.ops import FlushGate


class CompareEngine(BasicEngine):
    """
    CompareEngine is an engine which saves all commands. It is only intended
    for testing purposes. Two CompareEngine backends can be compared and
    return True if they contain the same commmands.
    """
    def __init__(self):
        BasicEngine.__init__(self)
        self._l = [[]]

    def is_available(self, cmd):
        return True

    def cache_cmd(self, cmd):
        # are there qubit ids that haven't been added to the list?
        all_qubit_id_list = [qubit.id for qureg in cmd.all_qubits
                             for qubit in qureg]
        maxidx = int(0)
        for qubit_id in all_qubit_id_list:
            maxidx = max(maxidx, qubit_id)

        # if so, increase size of list to account for these qubits
        add = maxidx+1-len(self._l)
        if add > 0:
            for i in range(add):
                self._l += [[]]

        # add gate command to each of the qubits involved
        for qubit_id in all_qubit_id_list:
            self._l[qubit_id] += [cmd]

    def receive(self, command_list):
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self.cache_cmd(cmd)
        if not self.is_last_engine:
            self.send(command_list)

    def compare_cmds(self, c1, c2):
        c2 = deepcopy(c2)
        c2.engine = c1.engine
        return c1 == c2

    def __eq__(self, other):
        if (not isinstance(other, CompareEngine) or
           len(self._l) != len(other._l)):
            return False
        for i in range(len(self._l)):
            if len(self._l[i]) != len(other._l[i]):
                return False
            for j in range(len(self._l[i])):
                if not self.compare_cmds(self._l[i][j], other._l[i][j]):
                    return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        string = ""
        for qubit_id in range(len(self._l)):
            string += "Qubit {0} : ".format(qubit_id)
            for command in self._l[qubit_id]:
                string += str(command) + ", "
            string = string[:-2] + "\n"
        return string


class DummyEngine(BasicEngine):
    """
    DummyEngine used for testing.

    The DummyEngine forwards all commands directly to next engine.
    If self.is_last_engine == True it just discards all gates.
    By setting save_commands == True all commands get saved as a
    list in self.received_commands. Elements are appended to this
    list so they are ordered according to when they are received.
    """
    def __init__(self, save_commands=False):
        """
        Initialize DummyEngine

        Args:
            save_commands (default = False): If True, commands are saved in
                                             self.received_commands.
        """
        BasicEngine.__init__(self)
        self.save_commands = save_commands
        self.received_commands = []

    def is_available(self, cmd):
        return True

    def receive(self, command_list):
        if self.save_commands:
            self.received_commands.extend(command_list)
        if not self.is_last_engine:
            self.send(command_list)
        else:
            pass



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition rule for the quantum Fourier transform.

Decomposes the QFT gate into Hadamard and controlled phase-shift gates (R).

Warning:
    The final Swaps are not included, as those are simply a re-indexing of
    quantum registers.
"""

import math

from projectq.cengines import DecompositionRule
from projectq.ops import H, R, QFT
from projectq.meta import Control


def _decompose_QFT(cmd):
    qb = cmd.qubits[0]
    eng = cmd.engine
    with Control(eng, cmd.control_qubits):
        for i in range(len(qb)):
            H | qb[-1 - i]
            for j in range(len(qb) - 1 - i):
                with Control(eng, qb[-1 - (j + i + 1)]):
                    R(math.pi / (1 << (1 + j))) | qb[-1 - i]


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(QFT.__class__, _decompose_QFT)
]



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Mapper for a quantum circuit to a 2D square grid.

Input: Quantum circuit with 1 and 2 qubit gates on n qubits. Gates are assumed
       to be applied in parallel if they act on disjoint qubit(s) and any pair
       of qubits can perform a 2 qubit gate (all-to-all connectivity)
Output: Quantum circuit in which qubits are placed in 2-D square grid in which
        only nearest neighbour qubits can perform a 2 qubit gate. The mapper
        uses Swap gates in order to move qubits next to each other.
"""
from copy import deepcopy
import itertools
import math
import random

import networkx as nx

from projectq.cengines import (BasicMapperEngine, LinearMapper,
                               return_swap_depth)
from projectq.meta import LogicalQubitIDTag
from projectq.ops import (AllocateQubitGate, Command, DeallocateQubitGate,
                          FlushGate, Swap)
from projectq.types import WeakQubitRef


class GridMapper(BasicMapperEngine):
    """
    Mapper to a 2-D grid graph.

    Mapped qubits on the grid are numbered in row-major order. E.g. for
    3 rows and 2 columns:

    0 - 1
    |   |
    2 - 3
    |   |
    4 - 5

    The numbers are the mapped qubit ids. The backend might number
    the qubits on the grid differently (e.g. not row-major), we call these
    backend qubit ids. If the backend qubit ids are not row-major, one can
    pass a dictionary translating from our row-major mapped ids to these
    backend ids.

    Note: The algorithm sorts twice inside each column and once inside each
          row.

    Attributes:
        current_mapping:  Stores the mapping: key is logical qubit id, value
                          is backend qubit id.
        storage(int): Number of gate it caches before mapping.
        num_rows(int): Number of rows in the grid
        num_columns(int): Number of columns in the grid
        num_qubits(int): num_rows x num_columns = number of qubits
        num_mappings (int): Number of times the mapper changed the mapping
        depth_of_swaps (dict): Key are circuit depth of swaps, value is the
                               number of such mappings which have been
                               applied
        num_of_swaps_per_mapping (dict): Key are the number of swaps per
                                         mapping, value is the number of such
                                         mappings which have been applied

    """
    def __init__(self, num_rows, num_columns, mapped_ids_to_backend_ids=None,
                 storage=1000,
                 optimization_function=lambda x: return_swap_depth(x),
                 num_optimization_steps=50):
        """
        Initialize a GridMapper compiler engine.

        Args:
            num_rows(int): Number of rows in the grid
            num_columns(int): Number of columns in the grid.
            mapped_ids_to_backend_ids(dict): Stores a mapping from mapped ids
                                             which are 0,...,self.num_qubits-1
                                             in row-major order on the grid to
                                             the corresponding qubit ids of the
                                             backend. Key: mapped id. Value:
                                             corresponding backend id.
                                             Default is None which means
                                             backend ids are identical to
                                             mapped ids.
            storage: Number of gates to temporarily store
            optimization_function: Function which takes a list of swaps and
                                   returns a cost value. Mapper chooses a
                                   permutation which minimizes this cost.
                                   Default optimizes for circuit depth.
            num_optimization_steps(int): Number of different permutations to
                                         of the matching to try and minimize
                                         the cost.
        Raises:
            RuntimeError: if incorrect `mapped_ids_to_backend_ids` parameter
        """
        BasicMapperEngine.__init__(self)
        self.num_rows = num_rows
        self.num_columns = num_columns
        self.num_qubits = num_rows * num_columns
        # Internally we use the mapped ids until sending a command.
        # Before sending we use this map to translate to backend ids:
        self._mapped_ids_to_backend_ids = mapped_ids_to_backend_ids
        if self._mapped_ids_to_backend_ids is None:
            self._mapped_ids_to_backend_ids = dict()
            for i in range(self.num_qubits):
                self._mapped_ids_to_backend_ids[i] = i
        if (not (set(self._mapped_ids_to_backend_ids.keys()) ==
                 set(list(range(self.num_qubits)))) or not (
                 len(set(self._mapped_ids_to_backend_ids.values())) ==
                 self.num_qubits)):
            raise RuntimeError("Incorrect mapped_ids_to_backend_ids parameter")
        self._backend_ids_to_mapped_ids = dict()
        for mapped_id, backend_id in self._mapped_ids_to_backend_ids.items():
            self._backend_ids_to_mapped_ids[backend_id] = mapped_id
        # As we use internally the mapped ids which are in row-major order,
        # we have an internal current mapping which maps from logical ids to
        # these mapped ids:
        self._current_row_major_mapping = deepcopy(self.current_mapping)
        self.storage = storage
        self.optimization_function = optimization_function
        self.num_optimization_steps = num_optimization_steps
        # Randomness to pick permutations if there are too many.
        # This creates an own instance of Random in order to not influence
        # the bound methods of the random module which might be used in other
        # places.
        self._rng = random.Random(11)
        # Storing commands
        self._stored_commands = list()
        # Logical qubit ids for which the Allocate gate has already been
        # processed and sent to the next engine but which are not yet
        # deallocated:
        self._currently_allocated_ids = set()
        # Change between 2D and 1D mappings (2D is a snake like 1D chain)
        # Note it translates to our mapped ids in row major order and not
        # backend ids which might be different.
        self._map_2d_to_1d = dict()
        self._map_1d_to_2d = dict()
        for row_index in range(self.num_rows):
            for column_index in range(self.num_columns):
                if row_index % 2 == 0:
                    mapped_id = row_index * self.num_columns + column_index
                    self._map_2d_to_1d[mapped_id] = mapped_id
                    self._map_1d_to_2d[mapped_id] = mapped_id
                else:
                    mapped_id_2d = row_index * self.num_columns + column_index
                    mapped_id_1d = ((row_index + 1) * self.num_columns -
                                    column_index - 1)
                    self._map_2d_to_1d[mapped_id_2d] = mapped_id_1d
                    self._map_1d_to_2d[mapped_id_1d] = mapped_id_2d
        # Statistics:
        self.num_mappings = 0
        self.depth_of_swaps = dict()
        self.num_of_swaps_per_mapping = dict()

    @property
    def current_mapping(self):
        return deepcopy(self._current_mapping)

    @current_mapping.setter
    def current_mapping(self, current_mapping):
        self._current_mapping = current_mapping
        if current_mapping is None:
            self._current_row_major_mapping = None
        else:
            self._current_row_major_mapping = dict()
            for logical_id, backend_id in current_mapping.items():
                self._current_row_major_mapping[logical_id] = (
                    self._backend_ids_to_mapped_ids[backend_id])

    def is_available(self, cmd):
        """
        Only allows 1 or two qubit gates.
        """
        num_qubits = 0
        for qureg in cmd.all_qubits:
            num_qubits += len(qureg)
        return (num_qubits <= 2)

    def _return_new_mapping(self):
        """
        Returns a new mapping of the qubits.

        It goes through self._saved_commands and tries to find a
        mapping to apply these gates on a first come first served basis.
        It reuses the function of a 1D mapper and creates a mapping for a
        1D linear chain and then wraps it like a snake onto the square grid.

        One might create better mappings by specializing this function for a
        square grid.

        Returns: A new mapping as a dict. key is logical qubit id,
                 value is mapped id
        """
        # Change old mapping to 1D in order to use LinearChain heuristic
        if self._current_row_major_mapping:
            old_mapping_1d = dict()
            for logical_id, mapped_id in (
                    self._current_row_major_mapping.items()):
                old_mapping_1d[logical_id] = self._map_2d_to_1d[mapped_id]
        else:
            old_mapping_1d = self._current_row_major_mapping

        new_mapping_1d = LinearMapper.return_new_mapping(
            num_qubits=self.num_qubits,
            cyclic=False,
            currently_allocated_ids=self._currently_allocated_ids,
            stored_commands=self._stored_commands,
            current_mapping=old_mapping_1d)

        new_mapping_2d = dict()
        for logical_id, mapped_id in new_mapping_1d.items():
            new_mapping_2d[logical_id] = self._map_1d_to_2d[mapped_id]
        return new_mapping_2d

    def _compare_and_swap(self, element0, element1, key):
        """
        If swapped (inplace), then return swap operation
        so that key(element0) < key(element1)
        """
        if key(element0) > key(element1):
            mapped_id0 = (element0.current_column +
                          element0.current_row * self.num_columns)
            mapped_id1 = (element1.current_column +
                          element1.current_row * self.num_columns)
            swap_operation = (mapped_id0, mapped_id1)
            # swap elements but update also current position:
            tmp_0 = element0.final_row
            tmp_1 = element0.final_column
            tmp_2 = element0.row_after_step_1
            element0.final_row = element1.final_row
            element0.final_column = element1.final_column
            element0.row_after_step_1 = element1.row_after_step_1
            element1.final_row = tmp_0
            element1.final_column = tmp_1
            element1.row_after_step_1 = tmp_2
            return swap_operation
        return None

    def _sort_within_rows(self, final_positions, key):
        swap_operations = []
        for row in range(self.num_rows):
            finished_sorting = False
            while not finished_sorting:
                finished_sorting = True
                for column in range(1, self.num_columns-1, 2):
                    element0 = final_positions[row][column]
                    element1 = final_positions[row][column+1]
                    swap = self._compare_and_swap(element0, element1, key=key)
                    if swap is not None:
                        finished_sorting = False
                        swap_operations.append(swap)
                for column in range(0, self.num_columns-1, 2):
                    element0 = final_positions[row][column]
                    element1 = final_positions[row][column+1]
                    swap = self._compare_and_swap(element0, element1, key=key)
                    if swap is not None:
                        finished_sorting = False
                        swap_operations.append(swap)
        return swap_operations

    def _sort_within_columns(self, final_positions, key):
        swap_operations = []
        for column in range(self.num_columns):
            finished_sorting = False
            while not finished_sorting:
                finished_sorting = True
                for row in range(1, self.num_rows-1, 2):
                    element0 = final_positions[row][column]
                    element1 = final_positions[row+1][column]
                    swap = self._compare_and_swap(element0, element1, key=key)
                    if swap is not None:
                        finished_sorting = False
                        swap_operations.append(swap)
                for row in range(0, self.num_rows-1, 2):
                    element0 = final_positions[row][column]
                    element1 = final_positions[row+1][column]
                    swap = self._compare_and_swap(element0, element1, key=key)
                    if swap is not None:
                        finished_sorting = False
                        swap_operations.append(swap)
        return swap_operations

    def return_swaps(self, old_mapping, new_mapping, permutation=None):
        """
        Returns the swap operation to change mapping

        Args:
            old_mapping: dict: keys are logical ids and values are mapped
                         qubit ids
            new_mapping: dict: keys are logical ids and values are mapped
                         qubit ids
            permutation: list of int from 0, 1, ..., self.num_rows-1. It is
                         used to permute the found perfect matchings. Default
                         is None which keeps the original order.
        Returns:
            List of tuples. Each tuple is a swap operation which needs to be
            applied. Tuple contains the two mapped qubit ids for the Swap.
        """
        if permutation is None:
            permutation = list(range(self.num_rows))
        swap_operations = []

        class Position(object):
            """ Custom Container."""
            def __init__(self, current_row, current_column, final_row,
                         final_column, row_after_step_1=None):
                self.current_row = current_row
                self.current_column = current_column
                self.final_row = final_row
                self.final_column = final_column
                self.row_after_step_1 = row_after_step_1

        # final_positions contains info containers
        # final_position[i][j] contains info container with
        # current_row == i and current_column == j
        final_positions = [[None for i in range(self.num_columns)]
                           for j in range(self.num_rows)]
        # move qubits which are in both mappings
        used_mapped_ids = set()
        for logical_id in old_mapping:
            if logical_id in new_mapping:
                used_mapped_ids.add(new_mapping[logical_id])
                old_column = old_mapping[logical_id] % self.num_columns
                old_row = old_mapping[logical_id] // self.num_columns
                new_column = new_mapping[logical_id] % self.num_columns
                new_row = new_mapping[logical_id] // self.num_columns
                info_container = Position(current_row=old_row,
                                          current_column=old_column,
                                          final_row=new_row,
                                          final_column=new_column)
                final_positions[old_row][old_column] = info_container
        # exchange all remaining None with the not yet used mapped ids
        all_ids = set(range(self.num_qubits))
        not_used_mapped_ids = list(all_ids.difference(used_mapped_ids))
        not_used_mapped_ids = sorted(not_used_mapped_ids, reverse=True)
        for row in range(self.num_rows):
            for column in range(self.num_columns):
                if final_positions[row][column] is None:
                    mapped_id = not_used_mapped_ids.pop()
                    new_column = mapped_id % self.num_columns
                    new_row = mapped_id // self.num_columns
                    info_container = Position(current_row=row,
                                              current_column=column,
                                              final_row=new_row,
                                              final_column=new_column)
                    final_positions[row][column] = info_container
        assert len(not_used_mapped_ids) == 0
        # 1. Assign column_after_step_1 for each element
        # Matching contains the num_columns matchings
        matchings = [None for i in range(self.num_rows)]
        # Build bipartite graph. Nodes are the current columns numbered
        # (0, 1, ...) and the destination columns numbered with an offset of
        # self.num_columns (0 + offset, 1+offset, ...)
        graph = nx.Graph()
        offset = self.num_columns
        graph.add_nodes_from(range(self.num_columns), bipartite=0)
        graph.add_nodes_from(range(offset, offset + self.num_columns),
                             bipartite=1)
        # Add an edge to the graph from (i, j+offset) for every element
        # currently in column i which should go to column j for the new
        # mapping
        for row in range(self.num_rows):
            for column in range(self.num_columns):
                destination_column = final_positions[row][column].final_column
                if not graph.has_edge(column, destination_column + offset):
                    graph.add_edge(column, destination_column + offset)
                    # Keep manual track of multiple edges between nodes
                    graph[column][destination_column + offset]['num'] = 1
                else:
                    graph[column][destination_column + offset]['num'] += 1
        # Find perfect matching, remove those edges from the graph
        # and do it again:
        for i in range(self.num_rows):
            top_nodes = range(self.num_columns)
            matching = nx.bipartite.maximum_matching(graph, top_nodes)
            matchings[i] = matching
            # Remove all edges of the current perfect matching
            for node in range(self.num_columns):
                if graph[node][matching[node]]['num'] == 1:
                    graph.remove_edge(node, matching[node])
                else:
                    graph[node][matching[node]]['num'] -= 1
        # permute the matchings:
        tmp = deepcopy(matchings)
        for i in range(self.num_rows):
            matchings[i] = tmp[permutation[i]]
        # Assign row_after_step_1
        for column in range(self.num_columns):
            for row_after_step_1 in range(self.num_rows):
                dest_column = matchings[row_after_step_1][column] - offset
                best_element = None
                for row in range(self.num_rows):
                    element = final_positions[row][column]
                    if element.row_after_step_1 is not None:
                        continue

                    if element.final_column == dest_column:
                        if best_element is None:
                            best_element = element
                        elif best_element.final_row > element.final_row:
                            best_element = element
                best_element.row_after_step_1 = row_after_step_1
        # 2. Sort inside all the rows
        swaps = self._sort_within_columns(final_positions=final_positions,
                                          key=lambda x: x.row_after_step_1)
        swap_operations += swaps
        # 3. Sort inside all the columns
        swaps = self._sort_within_rows(final_positions=final_positions,
                                       key=lambda x: x.final_column)
        swap_operations += swaps
        # 4. Sort inside all the rows
        swaps = self._sort_within_columns(final_positions=final_positions,
                                          key=lambda x: x.final_row)
        swap_operations += swaps
        return swap_operations

    def _send_possible_commands(self):
        """
        Sends the stored commands possible without changing the mapping.

        Note: self._current_row_major_mapping (hence also self.current_mapping)
              must exist already
        """
        active_ids = deepcopy(self._currently_allocated_ids)
        for logical_id in self._current_row_major_mapping:
            # So that loop doesn't stop before AllocateGate applied
            active_ids.add(logical_id)

        new_stored_commands = []
        for i in range(len(self._stored_commands)):
            cmd = self._stored_commands[i]
            if len(active_ids) == 0:
                new_stored_commands += self._stored_commands[i:]
                break
            if isinstance(cmd.gate, AllocateQubitGate):
                if cmd.qubits[0][0].id in self._current_row_major_mapping:
                    self._currently_allocated_ids.add(cmd.qubits[0][0].id)

                    mapped_id = self._current_row_major_mapping[
                        cmd.qubits[0][0].id]
                    qb = WeakQubitRef(
                        engine=self,
                        idx=self._mapped_ids_to_backend_ids[mapped_id])
                    new_cmd = Command(
                        engine=self,
                        gate=AllocateQubitGate(),
                        qubits=([qb],),
                        tags=[LogicalQubitIDTag(cmd.qubits[0][0].id)])
                    self.send([new_cmd])
                else:
                    new_stored_commands.append(cmd)
            elif isinstance(cmd.gate, DeallocateQubitGate):
                if cmd.qubits[0][0].id in active_ids:
                    mapped_id = self._current_row_major_mapping[
                        cmd.qubits[0][0].id]
                    qb = WeakQubitRef(
                        engine=self,
                        idx=self._mapped_ids_to_backend_ids[mapped_id])
                    new_cmd = Command(
                        engine=self,
                        gate=DeallocateQubitGate(),
                        qubits=([qb],),
                        tags=[LogicalQubitIDTag(cmd.qubits[0][0].id)])
                    self._currently_allocated_ids.remove(cmd.qubits[0][0].id)
                    active_ids.remove(cmd.qubits[0][0].id)
                    self._current_row_major_mapping.pop(cmd.qubits[0][0].id)
                    self._current_mapping.pop(cmd.qubits[0][0].id)
                    self.send([new_cmd])
                else:
                    new_stored_commands.append(cmd)
            else:
                send_gate = True
                mapped_ids = set()
                for qureg in cmd.all_qubits:
                    for qubit in qureg:
                        if qubit.id not in active_ids:
                            send_gate = False
                            break
                        mapped_ids.add(
                            self._current_row_major_mapping[qubit.id])
                # Check that mapped ids are nearest neighbour on 2D grid
                if len(mapped_ids) == 2:
                    qb0, qb1 = sorted(list(mapped_ids))
                    send_gate = False
                    if qb1 - qb0 == self.num_columns:
                        send_gate = True
                    elif qb1 - qb0 == 1 and qb1 % self.num_columns != 0:
                        send_gate = True
                if send_gate:
                    # Note: This sends the cmd correctly with the backend ids
                    #       as it looks up the mapping in self.current_mapping
                    #       and not our internal mapping
                    #       self._current_row_major_mapping
                    self._send_cmd_with_mapped_ids(cmd)
                else:
                    for qureg in cmd.all_qubits:
                        for qubit in qureg:
                            active_ids.discard(qubit.id)
                    new_stored_commands.append(cmd)
        self._stored_commands = new_stored_commands

    def _run(self):
        """
        Creates a new mapping and executes possible gates.

        It first allocates all 0, ..., self.num_qubits-1 mapped qubit ids, if
        they are not already used because we might need them all for the
        swaps. Then it creates a new map, swaps all the qubits to the new map,
        executes all possible gates, and finally deallocates mapped qubit ids
        which don't store any information.
        """
        num_of_stored_commands_before = len(self._stored_commands)
        if not self.current_mapping:
            self.current_mapping = dict()
        else:
            self._send_possible_commands()
            if len(self._stored_commands) == 0:
                return
        new_row_major_mapping = self._return_new_mapping()
        # Find permutation of matchings with lowest cost
        swaps = None
        lowest_cost = None
        matchings_numbers = list(range(self.num_rows))
        if self.num_optimization_steps <= math.factorial(self.num_rows):
            permutations = itertools.permutations(matchings_numbers,
                                                  self.num_rows)
        else:
            permutations = []
            for _ in range(self.num_optimization_steps):
                permutations.append(self._rng.sample(matchings_numbers,
                                                     self.num_rows))
        for permutation in permutations:
            trial_swaps = self.return_swaps(
                old_mapping=self._current_row_major_mapping,
                new_mapping=new_row_major_mapping,
                permutation=permutation)
            if swaps is None:
                swaps = trial_swaps
                lowest_cost = self.optimization_function(trial_swaps)
            elif lowest_cost > self.optimization_function(trial_swaps):
                swaps = trial_swaps
                lowest_cost = self.optimization_function(trial_swaps)
        if swaps:  # first mapping requires no swaps
            # Allocate all mapped qubit ids (which are not already allocated,
            # i.e., contained in self._currently_allocated_ids)
            mapped_ids_used = set()
            for logical_id in self._currently_allocated_ids:
                mapped_ids_used.add(
                    self._current_row_major_mapping[logical_id])
            not_allocated_ids = set(range(self.num_qubits)).difference(
                mapped_ids_used)
            for mapped_id in not_allocated_ids:
                qb = WeakQubitRef(
                    engine=self,
                    idx=self._mapped_ids_to_backend_ids[mapped_id])
                cmd = Command(engine=self, gate=AllocateQubitGate(),
                              qubits=([qb],))
                self.send([cmd])
            # Send swap operations to arrive at new_mapping:
            for qubit_id0, qubit_id1 in swaps:
                q0 = WeakQubitRef(
                    engine=self,
                    idx=self._mapped_ids_to_backend_ids[qubit_id0])
                q1 = WeakQubitRef(
                    engine=self,
                    idx=self._mapped_ids_to_backend_ids[qubit_id1])
                cmd = Command(engine=self, gate=Swap, qubits=([q0], [q1]))
                self.send([cmd])
            # Register statistics:
            self.num_mappings += 1
            depth = return_swap_depth(swaps)
            if depth not in self.depth_of_swaps:
                self.depth_of_swaps[depth] = 1
            else:
                self.depth_of_swaps[depth] += 1
            if len(swaps) not in self.num_of_swaps_per_mapping:
                self.num_of_swaps_per_mapping[len(swaps)] = 1
            else:
                self.num_of_swaps_per_mapping[len(swaps)] += 1
            # Deallocate all previously mapped ids which we only needed for the
            # swaps:
            mapped_ids_used = set()
            for logical_id in self._currently_allocated_ids:
                mapped_ids_used.add(new_row_major_mapping[logical_id])
            not_needed_anymore = set(range(self.num_qubits)).difference(
                mapped_ids_used)
            for mapped_id in not_needed_anymore:
                qb = WeakQubitRef(
                    engine=self,
                    idx=self._mapped_ids_to_backend_ids[mapped_id])
                cmd = Command(engine=self, gate=DeallocateQubitGate(),
                              qubits=([qb],))
                self.send([cmd])
        # Change to new map:
        self._current_row_major_mapping = new_row_major_mapping
        new_mapping = dict()
        for logical_id, mapped_id in new_row_major_mapping.items():
            new_mapping[logical_id] = (
                self._mapped_ids_to_backend_ids[mapped_id])
        self.current_mapping = new_mapping
        # Send possible gates:
        self._send_possible_commands()
        # Check that mapper actually made progress
        if len(self._stored_commands) == num_of_stored_commands_before:
            raise RuntimeError("Mapper is potentially in an infinite loop. " +
                               "It is likely that the algorithm requires " +
                               "too many qubits. Increase the number of " +
                               "qubits for this mapper.")

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it until
        we do a mapping (FlushGate or Cache of stored commands is full).

        Args:
            command_list (list of Command objects): list of commands to
                receive.
        """
        for cmd in command_list:
            if isinstance(cmd.gate, FlushGate):
                while(len(self._stored_commands)):
                    self._run()
                self.send([cmd])
            else:
                self._stored_commands.append(cmd)
            # Storage is full: Create new map and send some gates away:
            if len(self._stored_commands) >= self.storage:
                self._run()



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Defines a setup allowing to compile code for the IBM quantum chips:
->Any 5 qubit devices
->the ibmq online simulator
->the melbourne 15 qubit device

It provides the `engine_list` for the `MainEngine' based on the requested
device.  Decompose the circuit into a Rx/Ry/Rz/H/CNOT gate set that will be
translated in the backend in the U1/U2/U3/CX gate set.
"""

from projectq.setups import restrictedgateset
from projectq.ops import (Rx, Ry, Rz, H, CNOT, Barrier)
from projectq.cengines import (LocalOptimizer, IBM5QubitMapper,
                               SwapAndCNOTFlipper, BasicMapperEngine,
                               GridMapper)
from projectq.backends._ibm._ibm_http_client import show_devices


def get_engine_list(token=None, device=None):
    # Access to the hardware properties via show_devices
    # Can also be extended to take into account gate fidelities, new available
    # gate, etc..
    devices = show_devices(token)
    ibm_setup = []
    if device not in devices:
        raise DeviceOfflineError('Error when configuring engine list: device '
                                 'requested for Backend not connected')
    if devices[device]['nq'] == 5:
        # The requested device is a 5 qubit processor
        # Obtain the coupling map specific to the device
        coupling_map = devices[device]['coupling_map']
        coupling_map = list2set(coupling_map)
        mapper = IBM5QubitMapper(coupling_map)
        ibm_setup = [
            mapper,
            SwapAndCNOTFlipper(coupling_map),
            LocalOptimizer(10)
        ]
    elif device == 'ibmq_qasm_simulator':
        # The 32 qubit online simulator doesn't need a specific mapping for
        # gates. Can also run wider gateset but this setup keep the
        # restrictedgateset setup for coherence
        mapper = BasicMapperEngine()
        # Note: Manual Mapper doesn't work, because its map is updated only if
        # gates are applied if gates in the register are not used, then it
        # will lead to state errors
        res = dict()
        for i in range(devices[device]['nq']):
            res[i] = i
        mapper.current_mapping = res
        ibm_setup = [mapper]
    elif device == 'ibmq_16_melbourne':
        # Only 15 qubits available on this ibmqx2 unit(in particular qubit 7
        # on the grid), therefore need custom grid mapping
        grid_to_physical = {
            0: 0,
            1: 1,
            2: 2,
            3: 3,
            4: 4,
            5: 5,
            6: 6,
            7: 15,
            8: 14,
            9: 13,
            10: 12,
            11: 11,
            12: 10,
            13: 9,
            14: 8,
            15: 7
        }
        coupling_map = devices[device]['coupling_map']
        coupling_map = list2set(coupling_map)
        ibm_setup = [
            GridMapper(2, 8, grid_to_physical),
            LocalOptimizer(5),
            SwapAndCNOTFlipper(coupling_map),
            LocalOptimizer(5)
        ]
    else:
        # If there is an online device not handled into ProjectQ it's not too
        # bad, the engine_list can be constructed manually with the
        # appropriate mapper and the 'coupling_map' parameter
        raise DeviceNotHandledError('Device not yet fully handled by ProjectQ')

    # Most IBM devices accept U1,U2,U3,CX gates.
    # Most gates need to be decomposed into a subset that is manually converted
    # in the backend (until the implementation of the U1,U2,U3)
    # available gates decomposable now for U1,U2,U3: Rx,Ry,Rz and H
    setup = restrictedgateset.get_engine_list(one_qubit_gates=(Rx, Ry, Rz, H),
                                              two_qubit_gates=(CNOT, ),
                                              other_gates=(Barrier, ))
    setup.extend(ibm_setup)
    return setup


class DeviceOfflineError(Exception):
    pass


class DeviceNotHandledError(Exception):
    pass


def list2set(coupling_list):
    result = []
    for el in coupling_list:
        result.append(tuple(el))
    return set(result)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from projectq.ops import (Allocate, Deallocate, DaggeredGate, get_inverse,
                          Measure, SqrtSwap, Swap, X, Z)


def to_latex(circuit, drawing_order=None, draw_gates_in_parallel=True):
    """
    Translates a given circuit to a TikZ picture in a Latex document.

    It uses a json-configuration file which (if it does not exist) is created
    automatically upon running this function for the first time. The config
    file can be used to determine custom gate sizes, offsets, etc.

    New gate options can be added under settings['gates'], using the gate
    class name string as a key. Every gate can have its own width, height, pre
    offset and offset.

    Example:
        .. code-block:: python

            settings['gates']['HGate'] = {'width': .5, 'offset': .15}

    The default settings can be acquired using the get_default_settings()
    function, and written using write_settings().

    Args:
        circuit (list): Each qubit line is a list of
            CircuitItem objects, i.e., in circuit[line].
        drawing_order (list): A list of qubit lines from which
            the gates to be read from
        draw_gates_in_parallel (bool): If gates should (False)
            or not (True) be parallel in the circuit

    Returns:
        tex_doc_str (string): Latex document string which can be compiled
            using, e.g., pdflatex.
    """
    try:
        FileNotFoundError
    except NameError:
        FileNotFoundError = IOError  # for Python2 compatibility

    try:
        with open('settings.json') as settings_file:
            settings = json.load(settings_file)
    except FileNotFoundError:
        settings = write_settings(get_default_settings())

    text = _header(settings)
    text += _body(circuit,
                  settings,
                  drawing_order,
                  draw_gates_in_parallel=draw_gates_in_parallel)
    text += _footer(settings)
    return text


def write_settings(settings):
    """
    Write all settings to a json-file.

    Args:
        settings (dict): Settings dict to write.
    """
    with open('settings.json', 'w') as settings_file:
        json.dump(settings, settings_file, sort_keys=True, indent=4)
    return settings


def get_default_settings():
    """
    Return the default settings for the circuit drawing function to_latex().

    Returns:
        settings (dict): Default circuit settings
    """
    settings = dict()
    settings['gate_shadow'] = True
    settings['lines'] = ({
        'style': 'very thin',
        'double_classical': True,
        'init_quantum': True,
        'double_lines_sep': .04
    })
    settings['gates'] = ({
        'HGate': {
            'width': .5,
            'offset': .3,
            'pre_offset': .1
        },
        'XGate': {
            'width': .35,
            'height': .35,
            'offset': .1
        },
        'SqrtXGate': {
            'width': .7,
            'offset': .3,
            'pre_offset': .1
        },
        'SwapGate': {
            'width': .35,
            'height': .35,
            'offset': .1
        },
        'SqrtSwapGate': {
            'width': .35,
            'height': .35,
            'offset': .1
        },
        'Rx': {
            'width': 1.,
            'height': .8,
            'pre_offset': .2,
            'offset': .3
        },
        'Ry': {
            'width': 1.,
            'height': .8,
            'pre_offset': .2,
            'offset': .3
        },
        'Rz': {
            'width': 1.,
            'height': .8,
            'pre_offset': .2,
            'offset': .3
        },
        'Ph': {
            'width': 1.,
            'height': .8,
            'pre_offset': .2,
            'offset': .3
        },
        'EntangleGate': {
            'width': 1.8,
            'offset': .2,
            'pre_offset': .2
        },
        'DeallocateQubitGate': {
            'height': .15,
            'offset': .2,
            'width': .2,
            'pre_offset': .1
        },
        'AllocateQubitGate': {
            'height': .15,
            'width': .2,
            'offset': .1,
            'pre_offset': .1,
            'draw_id': False,
            'allocate_at_zero': False
        },
        'MeasureGate': {
            'width': 0.75,
            'offset': .2,
            'height': .5,
            'pre_offset': .2
        }
    })
    settings['control'] = {'size': .1, 'shadow': False}
    return settings


def _header(settings):
    """
    Writes the Latex header using the settings file.

    The header includes all packages and defines all tikz styles.

    Returns:
        header (string): Header of the Latex document.
    """
    packages = ("\\documentclass{standalone}\n\\usepackage[margin=1in]"
                "{geometry}\n\\usepackage[hang,small,bf]{caption}\n"
                "\\usepackage{tikz}\n"
                "\\usepackage{braket}\n\\usetikzlibrary{backgrounds,shadows."
                "blur,fit,decorations.pathreplacing,shapes}\n\n")

    init = ("\\begin{document}\n"
            "\\begin{tikzpicture}[scale=0.8, transform shape]\n\n")

    gate_style = ("\\tikzstyle{basicshadow}=[blur shadow={shadow blur steps=8,"
                  " shadow xshift=0.7pt, shadow yshift=-0.7pt, shadow scale="
                  "1.02}]")

    if not (settings['gate_shadow'] or settings['control']['shadow']):
        gate_style = ""

    gate_style += "\\tikzstyle{basic}=[draw,fill=white,"
    if settings['gate_shadow']:
        gate_style += "basicshadow"
    gate_style += "]\n"

    gate_style += ("\\tikzstyle{operator}=[basic,minimum size=1.5em]\n"
                   "\\tikzstyle{phase}=[fill=black,shape=circle," +
                   "minimum size={}".format(settings['control']['size']) +
                   "cm,inner sep=0pt,outer sep=0pt,draw=black")
    if settings['control']['shadow']:
        gate_style += ",basicshadow"
    gate_style += ("]\n\\tikzstyle{none}=[inner sep=0pt,outer sep=-.5pt,"
                   "minimum height=0.5cm+1pt]\n"
                   "\\tikzstyle{measure}=[operator,inner sep=0pt,minimum " +
                   "height={}cm, minimum width={}cm]\n".format(
                       settings['gates']['MeasureGate']['height'],
                       settings['gates']['MeasureGate']['width']) +
                   "\\tikzstyle{xstyle}=[circle,basic,minimum height=")
    x_gate_radius = min(settings['gates']['XGate']['height'],
                        settings['gates']['XGate']['width'])
    gate_style += ("{x_rad}cm,minimum width={x_rad}cm,inner sep=-1pt,"
                   "{linestyle}]\n").format(
                       x_rad=x_gate_radius,
                       linestyle=settings['lines']['style'])
    if settings['gate_shadow']:
        gate_style += ("\\tikzset{\nshadowed/.style={preaction={transform "
                       "canvas={shift={(0.5pt,-0.5pt)}}, draw=gray, opacity="
                       "0.4}},\n}\n")
    gate_style += "\\tikzstyle{swapstyle}=["
    gate_style += "inner sep=-1pt, outer sep=-1pt, minimum width=0pt]\n"
    edge_style = ("\\tikzstyle{edgestyle}=[" + settings['lines']['style'] +
                  "]\n")

    return packages + init + gate_style + edge_style


def _body(circuit, settings, drawing_order=None, draw_gates_in_parallel=True):
    """
    Return the body of the Latex document, including the entire circuit in
    TikZ format.

    Args:
        circuit (list<list<CircuitItem>>): Circuit to draw.
        settings: Dictionary of settings to use for the TikZ image.
        drawing_order: A list of circuit wires from where to read
            one gate command.
        draw_gates_in_parallel: Are the gate/commands occupying a
            single time step in the circuit diagram? For example, False means
            that gates can be parallel in the circuit.

    Returns:
        tex_str (string): Latex string to draw the entire circuit.
    """
    code = []

    conv = _Circ2Tikz(settings, len(circuit))

    to_where = None
    if drawing_order is None:
        drawing_order = list(range(len(circuit)))
    else:
        to_where = 1

    for line in drawing_order:
        code.append(
            conv.to_tikz(line,
                         circuit,
                         end=to_where,
                         draw_gates_in_parallel=draw_gates_in_parallel))

    return "".join(code)


def _footer(settings):
    """
    Return the footer of the Latex document.

    Returns:
        tex_footer_str (string): Latex document footer.
    """
    return r"\n\n\end{tikzpicture}\n\end{document}"


class _Circ2Tikz(object):
    """
    The Circ2Tikz class takes a circuit (list of lists of CircuitItem objects)
    and turns them into Latex/TikZ code.

    It uses the settings dictionary for gate offsets, sizes, spacing, ...
    """
    def __init__(self, settings, num_lines):
        """
        Initialize a circuit to latex converter object.

        Args:
            settings (dict): Dictionary of settings to use for the TikZ image.
            num_lines (int): Number of qubit lines to use for the entire
                circuit.
        """
        self.settings = settings
        self.pos = [0.] * num_lines
        self.op_count = [0] * num_lines
        self.is_quantum = [settings['lines']['init_quantum']] * num_lines

    def to_tikz(self, line, circuit, end=None, draw_gates_in_parallel=True):
        """
        Generate the TikZ code for one line of the circuit up to a certain
        gate.

        It modifies the circuit to include only the gates which have not been
        drawn. It automatically switches to other lines if the gates on those
        lines have to be drawn earlier.

        Args:
            line (int): Line to generate the TikZ code for.
            circuit (list<list<CircuitItem>>): The circuit to draw.
            end (int): Gate index to stop at (for recursion).
            draw_gates_in_parallel (bool): True or False for how to place gates

        Returns:
            tikz_code (string): TikZ code representing the current qubit line
                and, if it was necessary to draw other lines, those lines as
                well.
        """
        if end is None:
            end = len(circuit[line])

        tikz_code = []

        cmds = circuit[line]
        for i in range(0, end):
            gate = cmds[i].gate
            lines = cmds[i].lines
            ctrl_lines = cmds[i].ctrl_lines

            all_lines = lines + ctrl_lines
            all_lines.remove(line)  # remove current line
            for l in all_lines:
                gate_idx = 0
                while not (circuit[l][gate_idx] == cmds[i]):
                    gate_idx += 1

                tikz_code.append(self.to_tikz(l, circuit, gate_idx))

                # we are taking care of gate 0 (the current one)
                circuit[l] = circuit[l][1:]

            all_lines = lines + ctrl_lines
            pos = max([
                self.pos[l] for l in range(min(all_lines),
                                           max(all_lines) + 1)
            ])
            for l in range(min(all_lines), max(all_lines) + 1):
                self.pos[l] = pos + self._gate_pre_offset(gate)

            connections = ""
            for l in all_lines:
                connections += self._line(self.op_count[l] - 1,
                                          self.op_count[l],
                                          line=l)
            add_str = ""
            if gate == X:
                # draw NOT-gate with controls
                add_str = self._x_gate(lines, ctrl_lines)
                # and make the target qubit quantum if one of the controls is
                if not self.is_quantum[lines[0]]:
                    if sum([self.is_quantum[i] for i in ctrl_lines]) > 0:
                        self.is_quantum[lines[0]] = True
            elif gate == Z and len(ctrl_lines) > 0:
                add_str = self._cz_gate(lines + ctrl_lines)
            elif gate == Swap:
                add_str = self._swap_gate(lines, ctrl_lines)
            elif gate == SqrtSwap:
                add_str = self._sqrtswap_gate(lines,
                                              ctrl_lines,
                                              daggered=False)
            elif gate == get_inverse(SqrtSwap):
                add_str = self._sqrtswap_gate(lines, ctrl_lines, daggered=True)
            elif gate == Measure:
                # draw measurement gate
                for l in lines:
                    op = self._op(l)
                    width = self._gate_width(Measure)
                    height = self._gate_height(Measure)
                    shift0 = .07 * height
                    shift1 = .36 * height
                    shift2 = .1 * width
                    add_str += ("\n\\node[measure,edgestyle] ({op}) at ({pos}"
                                ",-{line}) {{}};\n\\draw[edgestyle] ([yshift="
                                "-{shift1}cm,xshift={shift2}cm]{op}.west) to "
                                "[out=60,in=180] ([yshift={shift0}cm]{op}."
                                "center) to [out=0, in=120] ([yshift=-{shift1}"
                                "cm,xshift=-{shift2}cm]{op}.east);\n"
                                "\\draw[edgestyle] ([yshift=-{shift1}cm]{op}."
                                "center) to ([yshift=-{shift2}cm,xshift=-"
                                "{shift1}cm]{op}.north east);").format(
                                    op=op,
                                    pos=self.pos[l],
                                    line=l,
                                    shift0=shift0,
                                    shift1=shift1,
                                    shift2=shift2)
                    self.op_count[l] += 1
                    self.pos[l] += (self._gate_width(gate) +
                                    self._gate_offset(gate))
                    self.is_quantum[l] = False
            elif gate == Allocate:
                # draw 'begin line'
                add_str = r"\n\node[none] ({}) at ({},-{}) {{$\Ket{{0}}{}$}};"
                id_str = ""
                if self.settings['gates']['AllocateQubitGate']['draw_id']:
                    id_str = "^{{\\textcolor{{red}}{{{}}}}}".format(cmds[i].id)
                xpos = self.pos[line]
                try:
                    if (self.settings['gates']['AllocateQubitGate']
                            ['allocate_at_zero']):
                        self.pos[line] -= self._gate_pre_offset(gate)
                        xpos = self._gate_pre_offset(gate)
                except KeyError:
                    pass
                self.pos[line] = max(
                    xpos + self._gate_offset(gate) + self._gate_width(gate),
                    self.pos[line])
                add_str = add_str.format(self._op(line), xpos, line, id_str)
                self.op_count[line] += 1
                self.is_quantum[line] = self.settings['lines']['init_quantum']
            elif gate == Deallocate:
                # draw 'end of line'
                op = self._op(line)
                add_str = "\n\\node[none] ({}) at ({},-{}) {{}};"
                add_str = add_str.format(op, self.pos[line], line)
                yshift = str(self._gate_height(gate)) + "cm]"
                add_str += (
                    "\n\\draw ([yshift={yshift}{op}.center) edge "
                    "[edgestyle] ([yshift=-{yshift}{op}.center);").format(
                        op=op, yshift=yshift)
                self.op_count[line] += 1
                self.pos[line] += (self._gate_width(gate) +
                                   self._gate_offset(gate))
            else:
                # regular gate must draw the lines it does not act upon
                # if it spans multiple qubits
                add_str = self._regular_gate(gate, lines, ctrl_lines)
                for l in lines:
                    self.is_quantum[l] = True

            tikz_code.append(add_str)
            if not gate == Allocate:
                tikz_code.append(connections)

            if not draw_gates_in_parallel:
                for idx in range(len(self.pos)):
                    if idx != line:
                        self.pos[idx] = self.pos[line]

        circuit[line] = circuit[line][end:]
        return "".join(tikz_code)

    def _gate_name(self, gate):
        """
        Return the string representation of the gate.

        Tries to use gate.tex_str and, if that is not available, uses str(gate)
        instead.

        Args:
            gate: Gate object of which to get the name / latex representation.

        Returns:
            gate_name (string): Latex gate name.
        """
        try:
            name = gate.tex_str()
        except AttributeError:
            name = str(gate)
        return name

    def _sqrtswap_gate(self, lines, ctrl_lines, daggered):
        """
        Return the TikZ code for a Square-root Swap-gate.

        Args:
            lines (list<int>): List of length 2 denoting the target qubit of
                the Swap gate.
            ctrl_lines (list<int>): List of qubit lines which act as controls.
            daggered (bool): Show the daggered one if True.
        """
        assert (len(lines) == 2)  # sqrt swap gate acts on 2 qubits
        delta_pos = self._gate_offset(SqrtSwap)
        gate_width = self._gate_width(SqrtSwap)
        lines.sort()

        gate_str = ""
        for line in lines:
            op = self._op(line)
            w = "{}cm".format(.5 * gate_width)
            s1 = "[xshift=-{w},yshift=-{w}]{op}.center".format(w=w, op=op)
            s2 = "[xshift={w},yshift={w}]{op}.center".format(w=w, op=op)
            s3 = "[xshift=-{w},yshift={w}]{op}.center".format(w=w, op=op)
            s4 = "[xshift={w},yshift=-{w}]{op}.center".format(w=w, op=op)
            swap_style = "swapstyle,edgestyle"
            if self.settings['gate_shadow']:
                swap_style += ",shadowed"
            gate_str += ("\n\\node[swapstyle] ({op}) at ({pos},-{line}) {{}};"
                         "\n\\draw[{swap_style}] ({s1})--({s2});\n"
                         "\\draw[{swap_style}] ({s3})--({s4});").format(
                             op=op,
                             s1=s1,
                             s2=s2,
                             s3=s3,
                             s4=s4,
                             line=line,
                             pos=self.pos[line],
                             swap_style=swap_style)

        # add a circled 1/2
        midpoint = (lines[0] + lines[1]) / 2.
        pos = self.pos[lines[0]]
        op_mid = "line{}_gate{}".format('{}-{}'.format(*lines),
                                        self.op_count[lines[0]])
        gate_str += ("\n\\node[xstyle] ({op}) at ({pos},-{line})\
                {{\\scriptsize $\\frac{{1}}{{2}}{dagger}$}};"
                     ).format(op=op_mid, line=midpoint, pos=pos,
                              dagger=r'^{{\dagger}}' if daggered else '')

        # add two vertical lines to connect circled 1/2
        gate_str += "\n\\draw ({}) edge[edgestyle] ({});".format(
            self._op(lines[0]), op_mid)
        gate_str += "\n\\draw ({}) edge[edgestyle] ({});".format(
            op_mid, self._op(lines[1]))

        if len(ctrl_lines) > 0:
            for ctrl in ctrl_lines:
                gate_str += self._phase(ctrl, self.pos[lines[0]])
                if ctrl > lines[1] or ctrl < lines[0]:
                    closer_line = lines[0]
                    if ctrl > lines[1]:
                        closer_line = lines[1]
                    gate_str += self._line(ctrl, closer_line)

        all_lines = ctrl_lines + lines
        new_pos = self.pos[lines[0]] + delta_pos + gate_width
        for i in all_lines:
            self.op_count[i] += 1
        for i in range(min(all_lines), max(all_lines) + 1):
            self.pos[i] = new_pos
        return gate_str

    def _swap_gate(self, lines, ctrl_lines):
        """
        Return the TikZ code for a Swap-gate.

        Args:
            lines (list<int>): List of length 2 denoting the target qubit of
                the Swap gate.
            ctrl_lines (list<int>): List of qubit lines which act as controls.

        """
        assert (len(lines) == 2)  # swap gate acts on 2 qubits
        delta_pos = self._gate_offset(Swap)
        gate_width = self._gate_width(Swap)
        lines.sort()

        gate_str = ""
        for line in lines:
            op = self._op(line)
            w = "{}cm".format(.5 * gate_width)
            s1 = "[xshift=-{w},yshift=-{w}]{op}.center".format(w=w, op=op)
            s2 = "[xshift={w},yshift={w}]{op}.center".format(w=w, op=op)
            s3 = "[xshift=-{w},yshift={w}]{op}.center".format(w=w, op=op)
            s4 = "[xshift={w},yshift=-{w}]{op}.center".format(w=w, op=op)
            swap_style = "swapstyle,edgestyle"
            if self.settings['gate_shadow']:
                swap_style += ",shadowed"
            gate_str += ("\n\\node[swapstyle] ({op}) at ({pos},-{line}) {{}};"
                         "\n\\draw[{swap_style}] ({s1})--({s2});\n"
                         "\\draw[{swap_style}] ({s3})--({s4});").format(
                             op=op,
                             s1=s1,
                             s2=s2,
                             s3=s3,
                             s4=s4,
                             line=line,
                             pos=self.pos[line],
                             swap_style=swap_style)
        gate_str += self._line(lines[0], lines[1])

        if len(ctrl_lines) > 0:
            for ctrl in ctrl_lines:
                gate_str += self._phase(ctrl, self.pos[lines[0]])
                if ctrl > lines[1] or ctrl < lines[0]:
                    closer_line = lines[0]
                    if ctrl > lines[1]:
                        closer_line = lines[1]
                    gate_str += self._line(ctrl, closer_line)

        all_lines = ctrl_lines + lines
        new_pos = self.pos[lines[0]] + delta_pos + gate_width
        for i in all_lines:
            self.op_count[i] += 1
        for i in range(min(all_lines), max(all_lines) + 1):
            self.pos[i] = new_pos
        return gate_str

    def _x_gate(self, lines, ctrl_lines):
        """
        Return the TikZ code for a NOT-gate.

        Args:
            lines (list<int>): List of length 1 denoting the target qubit of
                the NOT / X gate.
            ctrl_lines (list<int>): List of qubit lines which act as controls.

        """
        assert (len(lines) == 1)  # NOT gate only acts on 1 qubit
        line = lines[0]
        delta_pos = self._gate_offset(X)
        gate_width = self._gate_width(X)
        op = self._op(line)
        gate_str = ("\n\\node[xstyle] ({op}) at ({pos},-{line}) {{}};\n\\draw"
                    "[edgestyle] ({op}.north)--({op}.south);\n\\draw"
                    "[edgestyle] ({op}.west)--({op}.east);").format(
                        op=op, line=line, pos=self.pos[line])

        if len(ctrl_lines) > 0:
            for ctrl in ctrl_lines:
                gate_str += self._phase(ctrl, self.pos[line])
                gate_str += self._line(ctrl, line)

        all_lines = ctrl_lines + [line]
        new_pos = self.pos[line] + delta_pos + gate_width
        for i in all_lines:
            self.op_count[i] += 1
        for i in range(min(all_lines), max(all_lines) + 1):
            self.pos[i] = new_pos
        return gate_str

    def _cz_gate(self, lines):
        """
        Return the TikZ code for an n-controlled Z-gate.

        Args:
            lines (list<int>): List of all qubits involved.
        """
        assert len(lines) > 1
        line = lines[0]
        delta_pos = self._gate_offset(Z)
        gate_width = self._gate_width(Z)
        gate_str = self._phase(line, self.pos[line])

        for ctrl in lines[1:]:
            gate_str += self._phase(ctrl, self.pos[line])
            gate_str += self._line(ctrl, line)

        new_pos = self.pos[line] + delta_pos + gate_width
        for i in lines:
            self.op_count[i] += 1
        for i in range(min(lines), max(lines) + 1):
            self.pos[i] = new_pos
        return gate_str

    def _gate_width(self, gate):
        """
        Return the gate width, using the settings (if available).

        Returns:
            gate_width (float): Width of the gate.
                (settings['gates'][gate_class_name]['width'])
        """
        if isinstance(gate, DaggeredGate):
            gate = gate._gate
        try:
            gates = self.settings['gates']
            gate_width = gates[gate.__class__.__name__]['width']
        except KeyError:
            gate_width = .5
        return gate_width

    def _gate_pre_offset(self, gate):
        """
        Return the offset to use before placing this gate.

        Returns:
            gate_pre_offset (float): Offset to use before the gate.
                (settings['gates'][gate_class_name]['pre_offset'])
        """
        if isinstance(gate, DaggeredGate):
            gate = gate._gate
        try:
            gates = self.settings['gates']
            delta_pos = gates[gate.__class__.__name__]['pre_offset']
        except KeyError:
            delta_pos = self._gate_offset(gate)
        return delta_pos

    def _gate_offset(self, gate):
        """
        Return the offset to use after placing this gate and, if no pre_offset
        is defined, the same offset is used in front of the gate.

        Returns:
            gate_offset (float): Offset.
                (settings['gates'][gate_class_name]['offset'])
        """
        if isinstance(gate, DaggeredGate):
            gate = gate._gate
        try:
            gates = self.settings['gates']
            delta_pos = gates[gate.__class__.__name__]['offset']
        except KeyError:
            delta_pos = .2
        return delta_pos

    def _gate_height(self, gate):
        """
        Return the height to use for this gate.

        Returns:
            gate_height (float): Height of the gate.
                (settings['gates'][gate_class_name]['height'])
        """
        if isinstance(gate, DaggeredGate):
            gate = gate._gate
        try:
            height = self.settings['gates'][gate.__class__.__name__]['height']
        except KeyError:
            height = .5
        return height

    def _phase(self, line, pos):
        """
        Places a phase / control circle on a qubit line at a given position.

        Args:
            line (int): Qubit line at which to place the circle.
            pos (float): Position at which to place the circle.

        Returns:
            tex_str (string): Latex string representing a control circle at the
                given position.
        """
        phase_str = "\n\\node[phase] ({}) at ({},-{}) {{}};"
        return phase_str.format(self._op(line), pos, line)

    def _op(self, line, op=None, offset=0):
        """
        Returns the gate name for placing a gate on a line.

        Args:
            line (int): Line number.
            op (int): Operation number or, by default, uses the current op
                count.

        Returns:
            op_str (string): Gate name.
        """
        if op is None:
            op = self.op_count[line]
        return "line{}_gate{}".format(line, op + offset)

    def _line(self, p1, p2, double=False, line=None):
        """
        Connects p1 and p2, where p1 and p2 are either to qubit line indices,
        in which case the two most recent gates are connected, or two gate
        indices, in which case line denotes the line number and the two gates
        are connected on the given line.

        Args:
            p1 (int): Index of the first object to connect.
            p2 (int): Index of the second object to connect.
            double (bool): Draws double lines if True.
            line (int or None): Line index - if provided, p1 and p2 are gate
                indices.

        Returns:
            tex_str (string): Latex code to draw this / these line(s).
        """
        dbl_classical = self.settings['lines']['double_classical']

        if line is None:
            quantum = not dbl_classical or self.is_quantum[p1]
            op1, op2 = self._op(p1), self._op(p2)
            loc1, loc2 = 'north', 'south'
            shift = "xshift={}cm"
        else:
            quantum = not dbl_classical or self.is_quantum[line]
            op1, op2 = self._op(line, p1), self._op(line, p2)
            loc1, loc2 = 'west', 'east'
            shift = "yshift={}cm"

        if quantum:
            return "\n\\draw ({}) edge[edgestyle] ({});".format(op1, op2)

        if p2 > p1:
            loc1, loc2 = loc2, loc1
        edge_str = ("\n\\draw ([{shift}]{op1}.{loc1}) edge[edgestyle] "
                    "([{shift}]{op2}.{loc2});")
        line_sep = self.settings['lines']['double_lines_sep']
        shift1 = shift.format(line_sep / 2.)
        shift2 = shift.format(-line_sep / 2.)
        edges_str = edge_str.format(shift=shift1,
                                    op1=op1,
                                    op2=op2,
                                    loc1=loc1,
                                    loc2=loc2)
        edges_str += edge_str.format(shift=shift2,
                                     op1=op1,
                                     op2=op2,
                                     loc1=loc1,
                                     loc2=loc2)
        return edges_str

    def _regular_gate(self, gate, lines, ctrl_lines):
        """
        Draw a regular gate.

        Args:
            gate: Gate to draw.
            lines (list<int>): Lines the gate acts on.
            ctrl_lines (list<int>): Control lines.

        Returns:
            tex_str (string): Latex string drawing a regular gate at the given
                location
        """
        imax = max(lines)
        imin = min(lines)

        gate_lines = lines + ctrl_lines

        delta_pos = self._gate_offset(gate)
        gate_width = self._gate_width(gate)
        gate_height = self._gate_height(gate)

        name = self._gate_name(gate)

        lines = list(range(imin, imax + 1))

        tex_str = ""
        pos = self.pos[lines[0]]

        node_str = "\n\\node[none] ({}) at ({},-{}) {{}};"
        for l in lines:
            node1 = node_str.format(self._op(l), pos, l)
            node2 = ("\n\\node[none,minimum height={}cm,outer sep=0] ({}) at"
                     " ({},-{}) {{}};").format(gate_height,
                                               self._op(l, offset=1),
                                               pos + gate_width / 2., l)
            node3 = node_str.format(self._op(l, offset=2), pos + gate_width, l)
            tex_str += node1 + node2 + node3
            if l not in gate_lines:
                tex_str += self._line(self.op_count[l] - 1,
                                      self.op_count[l],
                                      line=l)

        tex_str += ("\n\\draw[operator,edgestyle,outer sep={width}cm] (["
                    "yshift={half_height}cm]{op1}) rectangle ([yshift=-"
                    "{half_height}cm]{op2}) node[pos=.5] {{{name}}};").format(
                        width=gate_width,
                        op1=self._op(imin),
                        op2=self._op(imax, offset=2),
                        half_height=.5 * gate_height,
                        name=name)

        for l in lines:
            self.pos[l] = pos + gate_width / 2.
            self.op_count[l] += 1

        for ctrl in ctrl_lines:
            if ctrl not in lines:
                tex_str += self._phase(ctrl, pos + gate_width / 2.)
                connect_to = imax
                if abs(connect_to - ctrl) > abs(imin - ctrl):
                    connect_to = imin
                tex_str += self._line(ctrl, connect_to)
                self.pos[ctrl] = pos + delta_pos + gate_width
                self.op_count[ctrl] += 1

        for l in lines:
            self.op_count[l] += 2

        for l in range(min(ctrl_lines + lines), max(ctrl_lines + lines) + 1):
            self.pos[l] = pos + delta_pos + gate_width
        return tex_str



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers decomposition for StatePreparation.
"""

import cmath
import math

from projectq.cengines import DecompositionRule
from projectq.meta import Control, Dagger
from projectq.ops import (StatePreparation, UniformlyControlledRy,
                          UniformlyControlledRz, Ph)


def _decompose_state_preparation(cmd):
    """
    Implements state preparation based on arXiv:quant-ph/0407010v1.
    """
    eng = cmd.engine
    assert len(cmd.qubits) == 1
    num_qubits = len(cmd.qubits[0])
    qureg = cmd.qubits[0]
    final_state = cmd.gate.final_state
    if len(final_state) != 2**num_qubits:
        raise ValueError("Length of final_state is invalid.")
    norm = 0.
    for amplitude in final_state:
        norm += abs(amplitude)**2
    if norm < 1 - 1e-10 or norm > 1 + 1e-10:
        raise ValueError("final_state is not normalized.")
    with Control(eng, cmd.control_qubits):
        # As in the paper reference, we implement the inverse:
        with Dagger(eng):
            # Cancel all the relative phases
            phase_of_blocks = []
            for amplitude in final_state:
                phase_of_blocks.append(cmath.phase(amplitude))
            for target_qubit in range(len(qureg)):
                angles = []
                phase_of_next_blocks = []
                for block in range(2**(len(qureg)-target_qubit-1)):
                    phase0 = phase_of_blocks[2*block]
                    phase1 = phase_of_blocks[2*block+1]
                    angles.append(phase0 - phase1)
                    phase_of_next_blocks.append((phase0 + phase1)/2.)
                UniformlyControlledRz(angles) | (qureg[(target_qubit+1):],
                                                 qureg[target_qubit])
                phase_of_blocks = phase_of_next_blocks
            # Cancel global phase
            Ph(-phase_of_blocks[0]) | qureg[-1]
            # Remove amplitudes from states which contain a bit value 1:
            abs_of_blocks = []
            for amplitude in final_state:
                abs_of_blocks.append(abs(amplitude))
            for target_qubit in range(len(qureg)):
                angles = []
                abs_of_next_blocks = []
                for block in range(2**(len(qureg)-target_qubit-1)):
                    a0 = abs_of_blocks[2*block]
                    a1 = abs_of_blocks[2*block+1]
                    if a0 == 0 and a1 == 0:
                        angles.append(0)
                    else:
                        angles.append(
                            -2. * math.acos(a0 / math.sqrt(a0**2 + a1**2)))
                    abs_of_next_blocks.append(math.sqrt(a0**2 + a1**2))
                UniformlyControlledRy(angles) | (qureg[(target_qubit+1):],
                                                 qureg[target_qubit])
                abs_of_blocks = abs_of_next_blocks


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(StatePreparation, _decompose_state_preparation)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from projectq.ops import BasicGate

from ._utils import _exec


class ControlFunctionOracle:
    """
    Synthesizes a negation controlled by an arbitrary control function.

    This creates a circuit for a NOT gate which is controlled by an arbitrary
    Boolean control function.  The control function is provided as integer
    representation of the function's truth table in binary notation.  For
    example, for the majority-of-three function, which truth table 11101000,
    the value for function can be, e.g., ``0b11101000``, ``0xe8``, or ``232``.

    Example:

        This example creates a circuit that causes to invert qubit ``d``,
        the majority-of-three function evaluates to true for the control
        qubits ``a``, ``b``, and ``c``.

        .. code-block:: python

            ControlFunctionOracle(0x8e) | ([a, b, c], d)
    """

    def __init__(self, function, **kwargs):
        """
        Initializes a control function oracle.

        Args:
            function (int): Function truth table.

        Keyword Args:
            synth: A RevKit synthesis command which creates a reversible
                   circuit based on a truth table and requires no additional
                   ancillae (e.g., ``revkit.esopbs``).  Can also be a nullary
                   lambda that calls several RevKit commands.
                   **Default:** ``revkit.esopbs``
        """
        if isinstance(function, int):
            self.function = function
        else:
            try:
                import dormouse
                self.function = dormouse.to_truth_table(function)
            except ImportError:  # pragma: no cover
                raise RuntimeError(
                    "The dormouse library needs to be installed in order to "
                    "automatically compile Python code into functions.  Try "
                    "to install dormouse with 'pip install dormouse'."
                )
        self.kwargs = kwargs

        self._check_function()

    def __or__(self, qubits):
        """
        Applies control function to qubits (and synthesizes circuit).

        Args:
            qubits (tuple<Qureg>): Qubits to which the control function is
                                   being applied. The first `n` qubits are for
                                   the controls, the last qubit is for the
                                   target qubit.
        """
        try:
            import revkit
        except ImportError:  # pragma: no cover
            raise RuntimeError(
                "The RevKit Python library needs to be installed and in the "
                "PYTHONPATH in order to call this function")

        # convert qubits to tuple
        qs = []
        for item in BasicGate.make_tuple_of_qureg(qubits):
            qs += item if isinstance(item, list) else [item]

        # function truth table cannot be larger than number of control qubits
        # allow
        if 2**(2**(len(qs) - 1)) <= self.function:
            raise AttributeError(
                "Function truth table exceeds number of control qubits")

        # create truth table from function integer
        hex_length = max(2**(len(qs) - 1) // 4, 1)
        revkit.tt(table="{0:#0{1}x}".format(self.function, hex_length))

        # create reversible circuit from truth table
        self.kwargs.get("synth", revkit.esopbs)()

        # check whether circuit has correct signature
        if revkit.ps(mct=True, silent=True)['qubits'] != len(qs):
            raise RuntimeError("Generated circuit lines does not match "
                               "provided qubits")

        # convert reversible circuit to ProjectQ code and execute it
        _exec(revkit.to_projectq(mct=True), qs)

    def _check_function(self):
        """
        Checks whether function is valid.
        """
        # function must be positive. We check in __or__ whether function is
        # too large
        if self.function < 0:
            raise AttributeError("Function must be a postive integer")
